# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/drum_group.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 16628 bytes
from typing import cast

from ...base import clamp, depends, first, listenable_property, listens, listens_group
from ...live import action, liveobj_changed, liveobj_valid
from ..controls import ButtonControl
from ..display import Renderable
from ..skin import LiveObjSkinEntry
from .clipboard import ClipboardComponent
from .drum_group_scroll import DrumGroupScrollComponent
from .note_editor import PitchProvider
from .playable import PlayableComponent

DEFAULT_DRUM_TRANSLATION_CHANNEL = 15
BASE_DRUM_GROUP_NOTE = 36


class DrumGroupComponent(PlayableComponent, PitchProvider, Renderable):
    mute_button = ButtonControl(color=None)
    solo_button = ButtonControl(color=None)
    delete_button = ButtonControl(color=None)
    quantize_button = ButtonControl(color=None)
    copy_button = ButtonControl(color=None)

    @depends(set_pad_translations=None, target_track=None)
    def __init__(
        self,
        name="Drum_Group",
        translation_channel=DEFAULT_DRUM_TRANSLATION_CHANNEL,
        set_pad_translations=None,
        target_track=None,
        scroll_component_type=None,
        clipboard_component_type=None,
        *a,
        **k,
    ):
        (super().__init__)(a, name=name, **k)
        self._target_track = target_track
        self._translation_channel = translation_channel
        self._set_pad_translations = set_pad_translations
        scroll_component_type = scroll_component_type or DrumGroupScrollComponent
        self._drum_group_scroller = scroll_component_type(parent=self)
        clipboard_component_type = clipboard_component_type or DrumPadClipboardComponent
        self._clipboard = clipboard_component_type(parent=self)
        self.register_slot(
            self._clipboard,
            lambda _: self._set_control_pads_from_script(self._is_copying()),
            "has_content",
        )
        self._drum_group_device = None
        self._selected_drum_pad = None
        self._all_drum_pads = []
        self._assigned_drum_pads = []

    @listenable_property
    def clipboard(self):
        return self._clipboard

    @property
    def has_assigned_drum_pads(self):
        return self._assigned_drum_pads and liveobj_valid(first(self._assigned_drum_pads))

    @property
    def assigned_drum_pads(self):
        return self._assigned_drum_pads

    def set_matrix(self, matrix):
        super().set_matrix(matrix)
        self._update_assigned_drum_pads()
        self._create_and_set_pad_translations()
        if self._any_modifier_pressed() or self._is_copying():
            self._set_control_pads_from_script(True)

    def __getattr__(self, name):
        if name.startswith("set_"):
            if "scroll" in name:
                return getattr(self._drum_group_scroller, name)
        raise AttributeError

    def set_drum_group_device(self, drum_group_device):
        if liveobj_changed(drum_group_device, self._drum_group_device):
            self._drum_group_device = drum_group_device
            self._drum_group_scroller.set_drum_group_device(drum_group_device)
            self._clipboard.set_drum_group_device(self._drum_group_device)
            drum_group_view = drum_group_device.view if drum_group_device else None
            self._DrumGroupComponent__on_visible_drum_pads_changed.subject = drum_group_device
            self._DrumGroupComponent__on_selected_drum_pad_changed.subject = drum_group_view
            self._DrumGroupComponent__on_drum_pads_scroll_position_changed.subject = drum_group_view
            self._DrumGroupComponent__on_chains_changed.subject = drum_group_device
            self._update_drum_pad_listeners()
            self._update_selected_drum_pad()
            self._update_note_translations()
            super().update()
        if not liveobj_valid(self._drum_group_device):
            self._update_assigned_drum_pads()
            self._update_led_feedback()

    def set_copy_button(self, button):
        self.copy_button.set_control_element(button)
        self._clipboard.set_copy_button(button)

    def quantize_pitch(self, note):
        pass

    def delete_pitch(self, drum_pad):
        pass

    def select_drum_pad(self, drum_pad):
        pass

    def _on_matrix_pressed(self, button):
        pad = self._pad_for_button(button)
        if not (liveobj_valid(self._drum_group_device) and liveobj_valid(pad)):
            return
        pad_name = cast(str, pad.name)
        if self.mute_button.is_pressed:
            pad.mute = not pad.mute
            self.notify(self.notifications.DrumGroup.Pad.mute, pad_name, pad.mute)
        if self.solo_button.is_pressed:
            pad.solo = not pad.solo
        if self.quantize_button.is_pressed:
            button.color = "DrumGroup.PadAction"
            self.quantize_pitch(pad.note)
        if self._is_copying():
            button.color = "DrumGroup.PadAction"
            self._clipboard.copy_or_paste(pad)
        if self.delete_button.is_pressed:
            button.color = "DrumGroup.PadAction"
            self._do_delete_pad(pad, pad_name)
        if self.select_button.is_pressed:
            self._do_select_pad(pad, pad_name)
            super()._on_matrix_pressed(button)
        if self.mute_button.is_pressed or self.solo_button.is_pressed:
            self._update_led_feedback()

    def _do_delete_pad(self, pad, pad_name):
        if action.delete_notes_with_pitch(self._target_track.target_clip, pad.note):
            self.notify(self.notifications.DrumGroup.Pad.delete_notes, pad_name)
            self.delete_pitch(pad)
        else:
            pad.delete_all_chains()
            self.notify(self.notifications.DrumGroup.Pad.delete, pad_name)

    def _do_select_pad(self, pad, pad_name):
        if liveobj_changed(self._drum_group_device.view.selected_drum_pad, pad):
            self._drum_group_device.view.selected_drum_pad = pad
            self.notify(self.notifications.DrumGroup.Pad.select, pad_name)
            self.select_drum_pad(pad)

    @mute_button.value
    def mute_button(self, _, button):
        self._set_control_pads_from_script(button.is_pressed)

    @solo_button.value
    def solo_button(self, _, button):
        self._set_control_pads_from_script(button.is_pressed)

    @delete_button.value
    def delete_button(self, _, button):
        self._set_control_pads_from_script(button.is_pressed)

    @copy_button.value
    def copy_button(self, *_):
        self._set_control_pads_from_script(self._is_copying())

    @quantize_button.value
    def quantize_button(self, _, button):
        self._set_control_pads_from_script(button.is_pressed)

    def _get_selected_drum_pad(self):
        if liveobj_valid(self._drum_group_device):
            return self._drum_group_device.view.selected_drum_pad
        return None

    def _is_copying(self):
        return self.copy_button.is_pressed or self._clipboard.has_content

    def _any_modifier_pressed(self):
        return any(
            [
                self.select_button.is_pressed,
                self.mute_button.is_pressed,
                self.solo_button.is_pressed,
                self.delete_button.is_pressed,
                self.copy_button.is_pressed,
                self.quantize_button.is_pressed,
            ],
        )

    def update(self):
        super().update()
        self._update_provided_pitches()

    def _update_provided_pitches(self):
        if liveobj_valid(self._selected_drum_pad):
            self.pitches = [self._selected_drum_pad.note]

    def _update_led_feedback(self):
        for button in self.matrix:
            self._update_button_color(button)

    def _update_button_color(self, button):
        pad = self._pad_for_button(button)
        button.color = self._color_for_pad(pad) if pad else "DrumGroup.PadEmpty"

    def _color_for_pad(self, pad):
        button_color = "DrumGroup.PadEmpty"
        is_selected = pad == self._selected_drum_pad
        if is_selected:
            button_color = "DrumGroup.PadSelected"
        if pad.chains:
            format_str = "Selected" if is_selected else ""
            if pad.mute:
                button_color = f"DrumGroup.PadMuted{format_str}"
            else:
                if pad.solo:
                    button_color = f"DrumGroup.PadSoloed{format_str}"
                else:
                    if not is_selected:
                        button_color = self._filled_color(pad)
            return button_color
        return None

    @staticmethod
    def _filled_color(pad):
        return LiveObjSkinEntry("DrumGroup.PadFilled", pad.chains[0])

    def _update_assigned_drum_pads(self):
        assigned_drum_pads = []
        visible_drum_pads = (
            self._drum_group_device.visible_drum_pads if liveobj_valid(self._drum_group_device) else None
        )
        if visible_drum_pads:
            if self._all_drum_pads:
                first_pad = first(visible_drum_pads)
                if first_pad:
                    size = self.width * self.height
                    first_note = first_pad.note
                    if first_note > 128 - size:
                        size = 128 - first_note
                    offset = clamp(first_note, 0, 128 - len(visible_drum_pads))
                    assigned_drum_pads = self._all_drum_pads[offset : offset + size]
        self._assigned_drum_pads = assigned_drum_pads

    def _update_drum_pad_listeners(self):
        if liveobj_valid(self._drum_group_device):
            self._all_drum_pads = self._drum_group_device.drum_pads
            visible_drum_pads = self._drum_group_device.visible_drum_pads
            self._DrumGroupComponent__on_solo_changed.replace_subjects(visible_drum_pads)
            self._DrumGroupComponent__on_mute_changed.replace_subjects(visible_drum_pads)
            self._DrumGroupComponent__on_color_changed.replace_subjects(
                [pad.chains[0] for pad in visible_drum_pads if pad.chains],
            )
            self._update_assigned_drum_pads()
            self._update_note_translations()

    def _update_selected_drum_pad(self):
        selected_drum_pad = self._get_selected_drum_pad()
        if liveobj_changed(self._selected_drum_pad, selected_drum_pad):
            self._selected_drum_pad = selected_drum_pad
            self._update_led_feedback()

    def _update_note_translations(self):
        if self._assigned_drum_pads:
            if not self._can_set_pad_translations():
                super()._update_note_translations()

    def _pad_for_button(self, button):
        if self.has_assigned_drum_pads:
            index = self._button_coordinates_to_pad_index(first(self._assigned_drum_pads).note, button.coordinate)
            if index < 128:
                return self._all_drum_pads[index]
            return None
        return None

    def _note_translation_for_button(self, button):
        identifier = None
        channel = None
        if self.has_assigned_drum_pads:
            identifier = self._button_coordinates_to_pad_index(first(self._assigned_drum_pads).note, button.coordinate)
            channel = self._translation_channel
        return (identifier, channel)

    def _button_coordinates_to_pad_index(self, first_note, coordinates):
        y, x = coordinates
        inverted_y = self.height - y - 1
        index = first_note + 4 * inverted_y + x
        if x >= 4:
            index += y * 4 + inverted_y * 4
        return index

    def _can_set_pad_translations(self):
        return self.width <= 4 and self.height <= 4

    def _create_and_set_pad_translations(self):
        def create_translation_entry(button):
            row, col = button.coordinate
            return (col, row, button.identifier, button.channel)

        if self._can_set_pad_translations():
            translations = []
            for button in self.matrix:
                button.channel = self._translation_channel
                button.identifier = self._button_coordinates_to_pad_index(BASE_DRUM_GROUP_NOTE, button.coordinate)
                button.enabled = True
                translations.append(create_translation_entry(button))

            self._set_pad_translations(tuple(translations))
        else:
            self._update_note_translations()
            self._set_pad_translations(None)

    @listens_group("solo")
    def __on_solo_changed(self, _):
        self._update_led_feedback()

    @listens_group("mute")
    def __on_mute_changed(self, _):
        self._update_led_feedback()

    @listens_group("color")
    def __on_color_changed(self, _):
        self._update_led_feedback()

    @listens("visible_drum_pads")
    def __on_visible_drum_pads_changed(self):
        self._update_drum_pad_listeners()
        self._update_led_feedback()

    @listens("drum_pads_scroll_position")
    def __on_drum_pads_scroll_position_changed(self):
        self._update_note_translations()
        self._update_led_feedback()

    @listens("selected_drum_pad")
    def __on_selected_drum_pad_changed(self):
        self._update_selected_drum_pad()
        self._update_provided_pitches()

    @listens("chains")
    def __on_chains_changed(self):
        self._update_led_feedback()


class DrumPadClipboardComponent(ClipboardComponent):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._drum_group_device = None

    def set_drum_group_device(self, drum_group):
        self._drum_group_device = drum_group
        self.clear()

    def _do_copy(self, obj):
        if not (liveobj_valid(obj) and obj.chains):
            self.notify(self.notifications.DrumGroup.Pad.CopyPaste.error_copy_from_empty_pad)
            return None
        self.notify(self.notifications.DrumGroup.Pad.CopyPaste.copy, cast(str, obj.name))
        return obj

    def _do_paste(self, obj):
        if liveobj_valid(obj):
            if self._source_obj.note != obj.note:
                if obj.chains:
                    obj.delete_all_chains()
                self._drum_group_device.copy_pad(self._source_obj.note, obj.note)
                self.notify(self.notifications.DrumGroup.Pad.CopyPaste.paste, cast(str, self._source_obj.name))
                return True
            self.notify(self.notifications.DrumGroup.Pad.CopyPaste.error_paste_to_source_pad)
        return False

    def _is_source_valid(self):
        return liveobj_valid(self._drum_group_device) and liveobj_valid(self._source_obj) and self._source_obj.chains



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/session.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 12515 bytes
from itertools import count

import Live

from ...base import EventObject, in_range, listens, listens_group, product
from ..component import Component
from ..control import ButtonControl
from .scene import SceneComponent


class SessionComponent(Component):
    _session_component_ends_initialisation = True
    scene_component_type = SceneComponent
    managed_select_button = ButtonControl(color="Session.Select", pressed_color="Session.SelectPressed")
    managed_delete_button = ButtonControl(color="Session.Delete", pressed_color="Session.DeletePressed")
    managed_duplicate_button = ButtonControl(color="Session.Duplicate", pressed_color="Session.DuplicatePressed")

    def __init__(self, session_ring=None, auto_name=False, *a, **k):
        (super().__init__)(*a, **k)
        self._session_ring = session_ring
        self._SessionComponent__on_offsets_changed.subject = self._session_ring
        self._stop_all_button = None
        self._stop_track_clip_buttons = None
        self._stop_clip_triggered_value = "Session.StopClipTriggered"
        self._stop_clip_value = "Session.StopClip"
        self._stop_clip_disabled_value = "Session.StopClipDisabled"
        self._track_slots = self.register_disconnectable(EventObject())
        self._selected_scene = self._create_scene()
        self._scenes = [self._create_scene() for _ in range(self._session_ring.num_scenes)]
        if self._session_component_ends_initialisation:
            self._end_initialisation()
        if auto_name:
            self._auto_name()
        self._SessionComponent__on_track_list_changed.subject = self.song
        self._SessionComponent__on_scene_list_changed.subject = self.song
        self._SessionComponent__on_selected_scene_changed.subject = self.song.view

    def _end_initialisation(self):
        self._SessionComponent__on_selected_scene_changed()
        self._reassign_scenes_and_tracks()

    def _create_scene(self):
        return self.scene_component_type(parent=self, session_ring=(self._session_ring))

    def scene(self, index):
        return self._scenes[index]

    def selected_scene(self):
        return self._selected_scene

    def _auto_name(self):
        self.name = "Session_Control"
        self.selected_scene().name = "Selected_Scene"
        for track_index in range(self._session_ring.num_tracks):
            clip_slot = self.selected_scene().clip_slot(track_index)
            clip_slot.name = "Selected_Scene_Clip_Slot_%d" % track_index

        for scene_index in range(self._session_ring.num_scenes):
            scene = self.scene(scene_index)
            scene.name = "Scene_%d" % scene_index
            for track_index in range(self._session_ring.num_tracks):
                clip_slot = scene.clip_slot(track_index)
                clip_slot.name = "%d_Clip_Slot_%d" % (track_index, scene_index)

    def set_stop_all_clips_button(self, button):
        self._stop_all_button = button
        self._SessionComponent__on_stop_all_value.subject = button
        self._update_stop_all_clips_button()

    def set_stop_track_clip_buttons(self, buttons):
        self._stop_track_clip_buttons = buttons
        self._SessionComponent__on_stop_track_value.replace_subjects(buttons or [])
        self._update_stop_track_clip_buttons()

    def set_managed_select_button(self, button):
        self.managed_select_button.set_control_element(button)
        self.set_modifier_button(button, "select")

    def set_managed_delete_button(self, button):
        self.managed_delete_button.set_control_element(button)
        self.set_modifier_button(button, "delete")

    def set_managed_duplicate_button(self, button):
        self.managed_duplicate_button.set_control_element(button)
        self.set_modifier_button(button, "duplicate")

    def set_modifier_button(self, button, name, clip_slots_only=False):
        for y in range(self._session_ring.num_scenes):
            scene = self.scene(y)
            if not clip_slots_only:
                getattr(scene, f"set_{name}_button")(button)
            else:
                for x in range(self._session_ring.num_tracks):
                    getattr(scene.clip_slot(x), f"set_{name}_button")(button)

    def set_clip_launch_buttons(self, buttons):
        if buttons:
            for button, (x, y) in buttons.iterbuttons():
                scene = self.scene(y)
                slot = scene.clip_slot(x)
                slot.set_launch_button(button)

        else:
            for x, y in product(range(self._session_ring.num_tracks), range(self._session_ring.num_scenes)):
                scene = self.scene(y)
                slot = scene.clip_slot(x)
                slot.set_launch_button(None)

    def set_scene_launch_buttons(self, buttons):
        if buttons:
            for x, button in enumerate(buttons):
                scene = self.scene(x)
                scene.set_launch_button(button)

        else:
            for index in range(self._session_ring.num_scenes):
                scene = self.scene(index)
                scene.set_launch_button(None)

    @listens("offset")
    def __on_offsets_changed(self, *a):
        if self.is_enabled():
            self._reassign_scenes_and_tracks()

    def _reassign_scenes_and_tracks(self):
        self._reassign_tracks()
        self._reassign_scenes()

    def set_rgb_mode(self, color_palette, color_table, clip_slots_only=False):
        for y in range(self._session_ring.num_scenes):
            scene = self.scene(y)
            if not clip_slots_only:
                scene.set_color_palette(color_palette)
                scene.set_color_table(color_table)
            else:
                for x in range(self._session_ring.num_tracks):
                    slot = scene.clip_slot(x)
                    slot.set_clip_palette(color_palette)
                    slot.set_clip_rgb_table(color_table)

    def update(self):
        super().update()
        if self.is_enabled():
            self._update_stop_track_clip_buttons()
            self._update_stop_all_clips_button()
            self._reassign_scenes_and_tracks()

    def _update_stop_track_clip_buttons(self):
        if self.is_enabled():
            for index in range(self._session_ring.num_tracks):
                self._update_stop_clips_led(index)

    @listens("scenes")
    def __on_scene_list_changed(self):
        self._reassign_scenes()

    @listens("visible_tracks")
    def __on_track_list_changed(self):
        self._reassign_tracks()

    @listens("selected_scene")
    def __on_selected_scene_changed(self):
        if self._selected_scene is not None:
            self._selected_scene.set_scene(self.song.view.selected_scene)

    def _update_stop_all_clips_button(self):
        if self.is_enabled():
            button = self._stop_all_button
            if button:
                button.set_light(button.is_pressed())

    def _reassign_scenes(self):
        scenes = self.song.scenes
        for index, scene in enumerate(self._scenes):
            scene_index = self._session_ring.scene_offset + index
            scene.set_scene(scenes[scene_index] if len(scenes) > scene_index else None)
            scene.set_track_offset(self._session_ring.track_offset)

        if self._selected_scene is not None:
            self._selected_scene.set_track_offset(self._session_ring.track_offset)

    def _reassign_tracks(self):
        tracks_to_use = self._session_ring.tracks_to_use()
        tracks = [t if isinstance(t, Live.Track.Track) else None for t in tracks_to_use]
        self._SessionComponent__on_fired_slot_index_changed.replace_subjects(tracks, count())
        self._SessionComponent__on_playing_slot_index_changed.replace_subjects(tracks, count())
        self._update_stop_all_clips_button()
        self._update_stop_track_clip_buttons()

    @listens("value")
    def __on_stop_all_value(self, value):
        self._stop_all_value(value)

    def _stop_all_value(self, value):
        if self.is_enabled():
            if not (value != 0 or self._stop_all_button.is_momentary()):
                self.song.stop_all_clips()
            self._update_stop_all_clips_button()

    @listens_group("value")
    def __on_stop_track_value(self, value, button):
        if self.is_enabled():
            if not (value != 0 or button.is_momentary()):
                tracks = self._session_ring.tracks_to_use()
                track_index = list(self._stop_track_clip_buttons).index(button) + self._session_ring.track_offset
                if in_range(track_index, 0, len(tracks)):
                    if tracks[track_index] in self.song.tracks:
                        tracks[track_index].stop_all_clips()

    @listens_group("fired_slot_index")
    def __on_fired_slot_index_changed(self, track_index):
        self._on_fired_slot_index_changed(track_index)

    def _on_fired_slot_index_changed(self, track_index):
        session_ring = self._session_ring
        button_index = track_index - session_ring.track_offset
        if in_range(button_index, 0, session_ring.num_tracks):
            self._update_stop_clips_led(button_index)

    @listens_group("playing_slot_index")
    def __on_playing_slot_index_changed(self, track_index):
        self._on_playing_slot_index_changed(track_index)

    def _on_playing_slot_index_changed(self, track_index):
        session_ring = self._session_ring
        button_index = track_index - session_ring.track_offset
        if in_range(button_index, 0, session_ring.num_tracks):
            self._update_stop_clips_led(button_index)

    def _update_stop_clips_led(self, index):
        tracks_to_use = self._session_ring.tracks_to_use()
        track_index = index + self._session_ring.track_offset
        if self.is_enabled():
            if self._stop_track_clip_buttons is not None:
                if index < len(self._stop_track_clip_buttons):
                    button = self._stop_track_clip_buttons[index]
                    if button is not None:
                        value_to_send = None
                        if track_index < len(tracks_to_use):
                            if tracks_to_use[track_index].clip_slots:
                                track = tracks_to_use[track_index]
                                if track.fired_slot_index == -2:
                                    value_to_send = self._stop_clip_triggered_value
                                else:
                                    if track.playing_slot_index >= 0:
                                        value_to_send = self._stop_clip_value
                                    else:
                                        value_to_send = self._stop_clip_disabled_value
                        if value_to_send is None:
                            button.set_light(False)
                        else:
                            if in_range(value_to_send, 0, 128):
                                button.send_value(value_to_send)
                            else:
                                button.set_light(value_to_send)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/accent.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 1513 bytes
from ...base import depends, listenable_property
from .. import Component
from ..controls import ToggleButtonControl
from ..display import Renderable


class AccentComponent(Component, Renderable):
    accent_button = ToggleButtonControl(color="Accent.Off", on_color="Accent.On")

    @depends(full_velocity=None)
    def __init__(self, name="Accent", full_velocity=None, *a, **k):
        (super().__init__)(a, name=name, **k)
        self._full_velocity = full_velocity
        self.accent_button.connect_property(self, "activated")

    @listenable_property
    def activated(self):
        return self._full_velocity.enabled

    @activated.setter
    def activated(self, state):
        if state != self._full_velocity.enabled:
            self._full_velocity.enabled = state
            self.notify(self.notifications.full_velocity, state)
            self.notify_activated()

    @accent_button.released_delayed
    def accent_button(self, _):
        self.activated = False



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/toggle.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 2802 bytes
from ...base import listens
from ..component import Component


class ToggleComponent(Component):
    is_private = True
    is_momentary = False
    read_only = False

    def __init__(
        self,
        property_name=None,
        subject=None,
        is_momentary=False,
        model_transform=None,
        view_transform=None,
        read_only=False,
        *a,
        **k,
    ):
        (super().__init__)(*a, **k)
        self._property_name = property_name
        self._property_slot = self.register_slot(subject, self._on_property_changed_in_model, property_name)
        self._property_button = None
        if is_momentary:
            self.is_momentary = is_momentary
        if model_transform:
            self.model_transform = model_transform
        if view_transform:
            self.view_transform = view_transform
        if read_only:
            self.read_only = read_only

    def model_transform(self, value):
        return value

    def view_transform(self, value):
        return value

    @property
    def subject(self):
        return self._property_slot.subject

    @subject.setter
    def subject(self, model):
        self._property_slot.subject = model
        self.update()

    @property
    def value(self):
        if self.subject:
            return getattr(self.subject, self._property_name)
        return False

    @value.setter
    def value(self, value):
        setattr(self.subject, self._property_name, value)

    def set_toggle_button(self, button):
        self._property_button = button
        self._ToggleComponent__on_button_value.subject = button
        self._update_button()

    def update(self):
        super().update()
        self._update_button()

    def _update_button(self):
        if self.is_enabled():
            button = self._property_button
            if button:
                button.set_light(self.view_transform(self.value))

    def _on_property_changed_in_model(self):
        self._update_button()

    @listens("value")
    def __on_button_value(self, value):
        if self.is_enabled() and not self.read_only:
            if self.is_momentary:
                if value:
                    self.value = self.model_transform(True)
                else:
                    self.value = self.model_transform(False)
        else:
            if not (value or self._property_button.is_momentary()):
                self.value = self.model_transform(not self.value)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/component.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 5392 bytes
import Live

from ..base import BooleanContext, depends, is_iterable, lazy_attribute, task
from .control import ControlManager


class Component(ControlManager):
    __events__ = ("enabled",)
    name = ""
    canonical_parent = None
    is_private = False
    _has_task_group = False
    _layer = None

    @depends(register_component=None, song=None)
    def __init__(self, name="", parent=None, register_component=None, song=None, layer=None, is_enabled=True, *a, **k):
        (super().__init__)(*a, **k)
        self.name = name
        self._parent = parent
        self._explicit_is_enabled = is_enabled
        self._recursive_is_enabled = True
        self._is_enabled = self._explicit_is_enabled
        self._song = song
        self._layer = layer
        self._child_components = []
        self._initializing_children = BooleanContext()
        if self._parent is not None:
            with self._initializing_children():
                self._parent.add_children(self)
        register_component(self)

    def disconnect(self):
        if self._has_task_group:
            self._tasks.kill()
            self._tasks.clear()
        super().disconnect()

    @property
    def parent(self):
        return self._parent

    @property
    def is_root(self):
        return self._parent is None

    def add_children(self, *a):
        components = list(map(self._add_child, a))
        if len(components) == 1:
            return components[0]
        return components

    def set_enabled(self, enable):
        self._explicit_is_enabled = bool(enable)
        self._update_is_enabled()
        for component in self._child_components:
            component._set_enabled_recursive(self.is_enabled())

    def is_enabled(self, explicit=False):
        if not explicit:
            return self._is_enabled
        return self._explicit_is_enabled

    def on_enabled_changed(self):
        self.update()

    def control_notifications_enabled(self):
        return self.is_enabled()

    @property
    def application(self):
        return Live.Application.get_application()

    @property
    def song(self):
        return self._song

    @property
    def layer(self):
        return self._layer

    @layer.setter
    def layer(self, new_layer):
        if self._layer != new_layer:
            self._release_all_layers()
            self._layer = new_layer
            if self.is_enabled():
                self._grab_all_layers()

    def _add_child(self, component):
        component._set_enabled_recursive(self.is_enabled())
        self._child_components.append(component)
        return component

    def _internal_on_enabled_changed(self):
        if self.is_enabled():
            self._grab_all_layers()
        else:
            self._release_all_layers()
        if self._has_task_group:
            if self.is_enabled():
                self._tasks.resume()
            else:
                self._tasks.pause()

    def _set_enabled_recursive(self, enable):
        self._recursive_is_enabled = bool(enable)
        self._update_is_enabled()
        for component in self._child_components:
            component._set_enabled_recursive(self.is_enabled())

    def _update_is_enabled(self):
        is_enabled = self._recursive_is_enabled and self._explicit_is_enabled
        if is_enabled != self._is_enabled:
            self._is_enabled = is_enabled
            self._internal_on_enabled_changed()
            if not self._initializing_children:
                self.on_enabled_changed()
                self.notify_enabled(is_enabled)

    @lazy_attribute
    @depends(parent_task_group=None)
    def _tasks(self, parent_task_group=None):
        tasks = parent_task_group.add(task.TaskGroup())
        if not self._is_enabled:
            tasks.pause()
        self._has_task_group = True
        return tasks

    def _grab_all_layers(self):
        for layer in self._get_layer_iterable():
            layer.grab(self)

    def _release_all_layers(self):
        for layer in self._get_layer_iterable():
            layer.release(self)

    def _get_layer_iterable(self):
        if self._layer is None:
            return ()
        if is_iterable(self._layer):
            return self._layer
        return (self._layer,)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/channel_strip.py
# Compiled at: 2023-12-21 15:35:35
# Size of source mod 2**32: 9693 bytes
from itertools import chain
from typing import cast

from ...base import find_if
from ...live import action, is_track_armed, liveobj_changed, liveobj_valid
from .. import Component
from ..controls import ButtonControl, MappedButtonControl, MappedControl, control_list
from ..display import Renderable
from ..skin import LiveObjSkinEntry

MAX_NUM_SENDS = 12
CROSSFADE_COLORS = ("Mixer.CrossfadeA", "Mixer.CrossfadeOff", "Mixer.CrossfadeB")


class ChannelStripComponent(Component, Renderable):
    include_in_top_level_state = False
    _active_instances = []

    @staticmethod
    def other_arm_buttons_pressed(strip):
        return (
            find_if(lambda x: x is not strip and x.arm_button.is_pressed, ChannelStripComponent._active_instances)
            is not None
        )

    @staticmethod
    def other_solo_buttons_pressed(strip):
        return (
            find_if(lambda x: x is not strip and x.solo_button.is_pressed, ChannelStripComponent._active_instances)
            is not None
        )

    volume_control = MappedControl()
    pan_control = MappedControl()
    send_controls = control_list(MappedControl, control_count=MAX_NUM_SENDS)
    indexed_send_controls = control_list(MappedControl, control_count=MAX_NUM_SENDS)
    track_select_button = ButtonControl(disabled_color="Mixer.NoTrack")
    mute_button = MappedButtonControl(disabled_color="Mixer.NoTrack", color="Mixer.MuteOff", on_color="Mixer.MuteOn")
    solo_button = ButtonControl(disabled_color="Mixer.NoTrack", color="Mixer.SoloOff", on_color="Mixer.SoloOn")
    arm_button = ButtonControl(disabled_color="Mixer.NoTrack", color="Mixer.ArmOff", on_color="Mixer.ArmOn")
    crossfade_cycle_button = ButtonControl(disabled_color="Mixer.NoTrack")
    shift_button = ButtonControl(color=None)

    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        ChannelStripComponent._active_instances.append(self)
        self._track = None
        self.register_slot(self.song.view, self._update_track_select_button, "selected_track")

        def make_property_slot(property_name, update_method):
            return self.register_slot(None, update_method, property_name)

        self._track_property_slots = [
            make_property_slot("solo", self._update_solo_button),
            make_property_slot("arm", self._update_arm_button),
            make_property_slot("implicit_arm", self._update_arm_button),
            make_property_slot("input_routing_type", self._update_arm_button),
            make_property_slot("color", self._update_track_select_button),
        ]
        self._mixer_device_property_slots = [
            make_property_slot("crossfade_assign", self._update_crossfade_cycle_button),
            make_property_slot("sends", self.update),
        ]

    def disconnect(self):
        ChannelStripComponent._active_instances.remove(self)
        self._track = None
        super().disconnect()

    @property
    def track(self):
        return self._track

    def set_track(self, track):
        if liveobj_changed(track, self._track):
            self._track = track if liveobj_valid(track) else None
            for slot in self._track_property_slots:
                slot.subject = track

            for slot in self._mixer_device_property_slots:
                slot.subject = track.mixer_device if track else None

            self.update()

    def set_indexed_send_control(self, control_element, index):
        if index < MAX_NUM_SENDS:
            self.indexed_send_controls.set_control_element_at_index(control_element, index)
            self.update()

    @track_select_button.pressed
    def track_select_button(self, _):
        if liveobj_changed(self.song.view.selected_track, self._track):
            self.song.view.selected_track = self._track
            self.notify(self.notifications.Track.select, cast(str, self._track.name))

    @solo_button.pressed
    def solo_button(self, _):
        solo_exclusive = (
            self.song.exclusive_solo != self.shift_button.is_pressed
            and not ChannelStripComponent.other_solo_buttons_pressed(self)
        )
        new_value = not self._track.solo
        for track in chain(self.song.tracks, self.song.return_tracks):
            if not (track == self._track or self._track).is_part_of_selection or track.is_part_of_selection:
                track.solo = new_value
            else:
                if solo_exclusive:
                    if track.solo:
                        track.solo = False

    @arm_button.pressed
    def arm_button(self, _):
        arm_exclusive = (
            self.song.exclusive_arm != self.shift_button.is_pressed
            and not ChannelStripComponent.other_arm_buttons_pressed(self)
        )
        action.arm((self._track), exclusive=arm_exclusive)

    @crossfade_cycle_button.pressed
    def crossfade_cycle_button(self, _):
        self._track.mixer_device.crossfade_assign = (self._track.mixer_device.crossfade_assign - 1) % len(
            self._track.mixer_device.crossfade_assignments.values,
        )

    def _all_controls(self):
        return chain([self.volume_control, self.pan_control], self.send_controls, self.indexed_send_controls)

    def update(self):
        super().update()
        if self.is_enabled():
            self._update_track_select_button()
            self._update_solo_button()
            self._update_arm_button()
            self._update_crossfade_cycle_button()
            if liveobj_valid(self._track):
                self._connect_parameters()
            else:
                self._disconnect_parameters()
        else:
            self._disconnect_parameters()

    def _disconnect_parameters(self):
        for control in self._all_controls():
            control.mapped_parameter = None

        self.mute_button.mapped_parameter = None

    def _connect_parameters(self):
        self.volume_control.mapped_parameter = self._track.mixer_device.volume
        self.pan_control.mapped_parameter = self._track.mixer_device.panning
        if self._track == self.song.master_track:
            self.mute_button.mapped_parameter = None
        else:
            self.mute_button.mapped_parameter = self._track.mixer_device.track_activator
        self._connect_send_parameters(self.send_controls)
        self._connect_send_parameters(self.indexed_send_controls)

    def _connect_send_parameters(self, send_controls):
        for index, send_control in enumerate(send_controls):
            if index < len(self._track.mixer_device.sends):
                send_control.mapped_parameter = self._track.mixer_device.sends[index]
            else:
                send_control.mapped_parameter = None

    def _update_track_select_button(self):
        has_track = liveobj_valid(self._track)
        self.track_select_button.enabled = has_track
        if has_track:
            is_selected = self.song.view.selected_track == self._track
            self.track_select_button.color = LiveObjSkinEntry(
                "Mixer.{}".format("Selected" if is_selected else "NotSelected"),
                self._track,
            )

    def _update_solo_button(self):
        track_valid = liveobj_valid(self._track) and self._track != self.song.master_track
        self.solo_button.enabled = track_valid
        if track_valid:
            self.solo_button.is_on = self._track.solo

    def _update_arm_button(self):
        track_valid = liveobj_valid(self._track) and self._track.can_be_armed
        self.arm_button.enabled = track_valid
        if track_valid:
            self.arm_button.on_color = "Mixer.ArmOn" if self._track.arm else "Mixer.ImplicitArmOn"
            self.arm_button.is_on = is_track_armed(self._track)

    def _update_crossfade_cycle_button(self):
        track_valid = liveobj_valid(self._track) and self._track != self.song.master_track
        self.crossfade_cycle_button.enabled = track_valid
        if track_valid:
            self.crossfade_cycle_button.color = CROSSFADE_COLORS[self._track.mixer_device.crossfade_assign]



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/combo.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 14975 bytes
from contextlib import contextmanager

from future.utils import string_types

from ...base import (
    PY3,
    EventObject,
    ProxyBase,
    const,
    depends,
    find_if,
    is_iterable,
    lazy_attribute,
    listens,
    liveobj_valid,
    nop,
    task,
)
from .. import defaults
from ..compound_element import CompoundElement
from ..control_element import NotifyingControlElement, get_element
from ..input_control_element import ParameterSlot
from ..resource import DEFAULT_PRIORITY
from .button import ButtonElementMixin


class WrapperElement(CompoundElement, ProxyBase):
    class ProxiedInterface(CompoundElement.ProxiedInterface):
        def __getattr__(self, name):
            if PY3:
                if "_wrapped_control" not in self.outer.__dict__:
                    raise AttributeError()
            wrapped = self.outer.__dict__["_wrapped_control"]
            return getattr(wrapped.proxied_interface, name)

    def __init__(self, wrapped_control=None, *a, **k):
        (super().__init__)(*a, **k)
        self._wrapped_control = get_element(wrapped_control)
        self._parameter_slot = ParameterSlot()

    @property
    def proxied_object(self):
        if self._is_initialized():
            if self.owns_control_element(self._wrapped_control):
                return self._wrapped_control
            return None
        return None

    def _is_initialized(self):
        return "_wrapped_control" in self.__dict__

    def register_wrapped(self):
        self.register_control_element(self._wrapped_control)

    def unregister_wrapped(self):
        self.unregister_control_element(self._wrapped_control)

    @property
    def wrapped_control(self):
        return self._wrapped_control

    def __bool__(self):
        return self.__nonzero__()

    def __nonzero__(self):
        return self.owns_control_element(self._wrapped_control)

    def on_nested_control_element_received(self, control):
        if control == self._wrapped_control:
            self._parameter_slot.control = control

    def on_nested_control_element_lost(self, control):
        if control == self._wrapped_control:
            self._parameter_slot.control = None

    def on_nested_control_element_value(self, value, control):
        if control == self._wrapped_control:
            self.notify_value(value)

    def connect_to(self, parameter):
        if not liveobj_valid(self._parameter_slot.parameter):
            self.request_listen_nested_control_elements()
        self._parameter_slot.parameter = parameter

    def release_parameter(self):
        if liveobj_valid(self._parameter_slot.parameter):
            self.unrequest_listen_nested_control_elements()
        self._parameter_slot.parameter = None


class ComboElement(WrapperElement):
    priority_increment = 0.5

    def __init__(self, control=None, modifier=None, *a, **k):
        if modifier is None:
            modifier = []
        (super().__init__)(a, wrapped_control=control, **k)
        self._combo_modifiers = (
            list(map(get_element, modifier))
            if isinstance(modifier, string_types) or is_iterable(modifier)
            else [get_element(modifier)]
        )
        (self.register_control_elements)(*self._combo_modifiers)
        self.request_listen_nested_control_elements()

    def reset(self):
        if self.owns_control_element(self._wrapped_control):
            self._wrapped_control.reset()

    def get_control_element_priority(self, element, priority):
        if element == self._wrapped_control:
            priority = DEFAULT_PRIORITY if priority is None else priority
            return priority + self.priority_increment
        return priority

    def on_nested_control_element_received(self, control):
        if control != self._wrapped_control:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_received(control)

    def on_nested_control_element_lost(self, control):
        if control != self._wrapped_control:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_lost(control)

    def on_nested_control_element_value(self, value, control):
        if control != self._wrapped_control:
            self._enforce_control_invariant()
        else:
            super().on_nested_control_element_value(value, control)

    def _enforce_control_invariant(self):
        if self._combo_is_on():
            if not self.has_control_element(self._wrapped_control):
                self.register_control_element(self._wrapped_control)
        else:
            if self.has_control_element(self._wrapped_control):
                self.unregister_control_element(self._wrapped_control)

    def _combo_is_on(self):
        return all(map(self._modifier_is_valid, self._combo_modifiers))

    def _modifier_is_valid(self, mod):
        return self.owns_control_element(mod) and mod.is_pressed()


class EventElement(NotifyingControlElement, ProxyBase, ButtonElementMixin):
    event_value = 1
    _subject = None

    def __init__(self, subject=None, event_name=None, *a, **k):
        (super().__init__)(*a, **k)
        self._subject = subject
        self.register_slot(subject, self._on_event, event_name)

    @property
    def proxied_object(self):
        return self._subject

    @property
    def proxied_interface(self):
        return getattr(self._subject, "proxied_interface", self._subject)

    def _on_event(self, *a, **k):
        self.notify_value(self.event_value)

    def is_momentary(self):
        return False

    def reset(self):
        getattr(self._subject, "reset", nop)()

    def send_value(self, *a, **k):
        try:
            send_value = super().__getattr__("send_value")
        except AttributeError:
            send_value = nop

        send_value(*a, **k)

    def set_light(self, *a, **k):
        try:
            set_light = super().__getattr__("set_light")
        except AttributeError:
            set_light = nop

        set_light(*a, **k)


class DoublePressContext(EventObject):
    __events__ = ("break_double_press",)

    @contextmanager
    def breaking_double_press(self):
        self._broke_double_press = False
        yield
        if not self._broke_double_press:
            self.break_double_press()

    def break_double_press(self):
        self.notify_break_double_press()
        self._broke_double_press = True


GLOBAL_DOUBLE_PRESS_CONTEXT_PROVIDER = const(DoublePressContext())


class DoublePressElement(WrapperElement):
    __events__ = ("single_press", "double_press")
    DOUBLE_PRESS_MAX_DELAY = defaults.MOMENTARY_DELAY

    @depends(double_press_context=GLOBAL_DOUBLE_PRESS_CONTEXT_PROVIDER)
    def __init__(self, wrapped_control=None, double_press_context=None, *a, **k):
        (super().__init__)(a, wrapped_control=wrapped_control, **k)
        self.register_control_element(self._wrapped_control)
        self._double_press_context = double_press_context
        self._double_press_task = self._tasks.add(
            task.sequence(task.wait(self.DOUBLE_PRESS_MAX_DELAY), task.run(self.finish_single_press)),
        ).kill()
        self.request_listen_nested_control_elements()

    def on_nested_control_element_value(self, value, control):
        if not control.is_momentary() or value and self._double_press_task.is_killed:
            self._double_press_context.break_double_press()
            self._DoublePressElement__on_break_double_press.subject = self._double_press_context
            self._double_press_task.restart()
        else:
            self.finish_double_press()
            self._double_press_task.kill()
        super().on_nested_control_element_value(value, control)

    @listens("break_double_press")
    def __on_break_double_press(self):
        if not self._double_press_task.is_killed:
            self._double_press_task.kill()
            self.finish_single_press()

    def finish_single_press(self):
        self._DoublePressElement__on_break_double_press.subject = None
        self.notify_single_press()

    def finish_double_press(self):
        self._DoublePressElement__on_break_double_press.subject = None
        self.notify_double_press()

    @lazy_attribute
    def single_press(self):
        return EventElement(self, "single_press")

    @lazy_attribute
    def double_press(self):
        return EventElement(self, "double_press")


class MultiElement(CompoundElement, ButtonElementMixin):
    class ProxiedInterface(CompoundElement.ProxiedInterface):
        def __getattr__(self, name):
            found = find_if(
                lambda x: x is not None,
                (getattr(c.proxied_interface, name, None) for c in self.outer.nested_control_elements()),
            )
            if found is not None:
                return found
            raise AttributeError

    def __init__(self, *controls, **k):
        (super().__init__)(**k)
        (self.register_control_elements)(*map(get_element, controls))

    def send_value(self, value):
        for control in self.owned_control_elements():
            control.send_value(value)

    def set_light(self, value):
        for control in self.owned_control_elements():
            control.set_light(value)

    def on_nested_control_element_value(self, value, control):
        if not self.is_pressed() or value:
            self.notify_value(value)

    def is_pressed(self):
        return find_if(lambda c: getattr(c, "is_pressed", const(False))(), self.owned_control_elements()) is not None

    def is_momentary(self):
        return find_if(lambda c: getattr(c, "is_momentary", const(False))(), self.nested_control_elements()) is not None

    def on_nested_control_element_received(self, control):
        pass

    def on_nested_control_element_lost(self, control):
        pass


class ToggleElement(WrapperElement):
    def __init__(self, on_control=None, off_control=None, *a, **k):
        (super().__init__)(*a, **k)
        self._on_control = get_element(on_control)
        self._off_control = get_element(off_control)
        self._toggled = False
        self._update_toggled()

    def set_toggled(self, value):
        self._toggled = value
        self._update_toggled()

    def _update_toggled(self):
        if self.has_control_element(self._wrapped_control):
            self.unregister_control_element(self._wrapped_control)
        self._wrapped_control = self._on_control if self._toggled else self._off_control
        if self._wrapped_control is not None:
            self.register_control_element(self._wrapped_control)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/page.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4703 bytes
from ...base import EventObject, clamp, listens
from .. import Component
from .scroll import Scrollable, ScrollComponent


class Pageable(EventObject):
    __events__ = ("position",)
    position_count = NotImplemented
    position = NotImplemented
    page_offset = NotImplemented
    page_length = NotImplemented


class PageComponent(Component, Scrollable):
    def __init__(self, pageable=None, scroll_skin_name=None, *a, **k):
        (super().__init__)(*a, **k)
        self._pageable = pageable or self
        self._position_scroll = ScrollComponent(self, parent=self, scroll_skin_name=scroll_skin_name)
        self._page_scroll = ScrollComponent(parent=self, scroll_skin_name=scroll_skin_name)
        self._page_scroll.can_scroll_up = self.can_scroll_page_up
        self._page_scroll.can_scroll_down = self.can_scroll_page_down
        self._page_scroll.scroll_down = self.scroll_page_down
        self._page_scroll.scroll_up = self.scroll_page_up
        self._PageComponent__on_position_changed.subject = self._pageable

    def set_scroll_up_button(self, button):
        self._position_scroll.set_scroll_up_button(button)

    def set_scroll_down_button(self, button):
        self._position_scroll.set_scroll_down_button(button)

    def set_scroll_encoder(self, encoder):
        self._position_scroll.set_scroll_encoder(encoder)

    def set_scroll_page_up_button(self, button):
        self._page_scroll.set_scroll_up_button(button)

    def set_scroll_page_down_button(self, button):
        self._page_scroll.set_scroll_down_button(button)

    def set_scroll_page_encoder(self, encoder):
        self._page_scroll.set_scroll_encoder(encoder)

    def scroll_page_up(self):
        self._scroll_page(1)

    def scroll_page_down(self):
        self._scroll_page(-1)

    def scroll_up(self):
        self._scroll_position(1)

    def scroll_down(self):
        self._scroll_position(-1)

    def can_scroll_page_up(self):
        model = self._pageable
        return model.position < model.position_count - model.page_length

    def can_scroll_page_down(self):
        return self._pageable.position > 0

    def can_scroll_up(self):
        return self.can_scroll_page_up()

    def can_scroll_down(self):
        return self.can_scroll_page_down()

    def _scroll_position(self, delta):
        if self.is_enabled():
            model = self._pageable
            model.position = clamp(model.position + delta, 0, model.position_count - model.page_length)

    def _scroll_page(self, sign):
        if self.is_enabled():
            model = self._pageable
            remainder = (model.position - model.page_offset) % model.page_length
            delta = model.page_length - remainder if sign > 0 else -model.page_length if remainder == 0 else -remainder
            self._scroll_position(delta)

    def update(self):
        super().update()
        self._position_scroll.update()
        self._page_scroll.update()

    @listens("position")
    def __on_position_changed(self):
        self._position_scroll.update()
        self._page_scroll.update()



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/skin.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3490 bytes
from itertools import chain
from typing import Any, NamedTuple, Optional

from . import BasicColors

ON_SUFFIXES = ("enabled", "on", "pressed", "selected")


class OptionalSkinEntry(NamedTuple):
    name: str
    fallback_name = None
    fallback_name: Optional[str]


class LiveObjSkinEntry(NamedTuple):
    name: str
    liveobj: Any


class Skin:
    def __init__(self, colors=None, *a, **k):
        (super().__init__)(*a, **k)
        self.colors = {}
        if colors is not None:
            self._fill_colors(colors)

    def _fill_colors(self, colors, pathname=""):
        if getattr(colors, "__bases__", None):
            for base in colors.__bases__:
                self._fill_colors(base, pathname=pathname)

        for k, v in vars(colors).items():
            if k[:1] != "_":
                if callable(v):
                    self._fill_colors(v, f"{pathname}{k}.")
                else:
                    self.colors[f"{pathname}{k}"] = v

    def __getitem__(self, key):
        key = self._from_wrapper(key)
        if key is None:
            return None
        if isinstance(key, LiveObjSkinEntry):
            key_name = self._from_wrapper(key.name)
            color = self.colors[key_name]
            if callable(color):
                return color(key.liveobj)
            key = key_name
        if key not in self.colors:
            if key.lower().endswith(ON_SUFFIXES):
                return BasicColors.ON
            return BasicColors.OFF
        return self.colors[key]

    def _from_wrapper(self, key):
        if isinstance(key, OptionalSkinEntry):
            if key.name in self.colors:
                return key.name
            return key.fallback_name
        return key


def merge_skins(*skins):
    skin = Skin()
    skin.colors = dict(chain(*(s.colors.items() for s in skins)))
    return skin



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/control/mapped.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 6273 bytes
from control_surface.internal_parameter import InternalParameterBase

from ...base import clamp, listens, liveobj_valid, old_hasattr
from .control import InputControl
from .encoder import ValueStepper


class MappedControl(InputControl):
    class State(InputControl.State):
        def __init__(self, control=None, manager=None, *a, **k):
            (super(MappedControl.State, self).__init__)(a, control=control, manager=manager, **k)
            self._direct_mapping = None

        def set_control_element(self, control_element):
            if self._control_element:
                self._control_element.release_parameter()
            super(MappedControl.State, self).set_control_element(control_element)
            self._update_direct_connection()
            self._update_value_slot()

        @property
        def mapped_parameter(self):
            return self._direct_mapping

        @mapped_parameter.setter
        def mapped_parameter(self, direct_mapping):
            self._direct_mapping = direct_mapping
            self._update_direct_connection()
            self._update_value_slot()

        def _event_listener_required(self):
            return True

        def _update_direct_connection(self):
            if self._control_element:
                self._control_element.connect_to(self._direct_mapping)

        def _update_value_slot(self):
            self._value_slot.subject = (
                None
                if (liveobj_valid(self._direct_mapping)) and (self._direct_mapping.is_enabled)
                else (self._control_element)
            )

        def _notifications_enabled(self):
            return super(MappedControl.State, self)._notifications_enabled() and self._direct_mapping is None

    def __init__(self, *a, **k):
        super().__init__(extra_args=a, extra_kws=k)


def is_internal_parameter(parameter):
    return isinstance(parameter, InternalParameterBase)


def is_zoom_parameter(parameter):
    return old_hasattr(parameter, "zoom")


class MappedSensitivitySettingControl(MappedControl):
    DEFAULT_SENSITIVITY = 1.0
    FINE_SENSITIVITY = 0.1

    class State(MappedControl.State):
        def __init__(self, *a, **k):
            (super(MappedSensitivitySettingControl.State, self).__init__)(*a, **k)
            self.default_sensitivity = MappedSensitivitySettingControl.DEFAULT_SENSITIVITY
            self.fine_sensitivity = MappedSensitivitySettingControl.FINE_SENSITIVITY
            self._quantized_stepper = ValueStepper()

        def update_sensitivities(self, default, fine):
            self.default_sensitivity = default
            self.fine_sensitivity = fine
            if self._control_element:
                self._update_control_sensitivity()

        def _update_direct_connection(self):
            if self._control_element is None or is_internal_parameter(self.mapped_parameter):
                if self._control_element:
                    self._control_element.release_parameter()
                self._control_value.subject = self._control_element
            else:
                self._control_value.subject = None
                self._update_control_element()
            self._quantized_stepper.reset()

        def _update_control_element(self):
            if liveobj_valid(self.mapped_parameter):
                self._control_element.connect_to(self.mapped_parameter)
            else:
                self._control_element.release_parameter()
            self._update_control_sensitivity()
            self._quantized_stepper.reset()

        def _update_control_sensitivity(self):
            if old_hasattr(self._control_element, "set_sensitivities"):
                self._control_element.set_sensitivities(self.default_sensitivity, self.fine_sensitivity)
            else:
                self._control_element.mapping_sensitivity = self.default_sensitivity

        @listens("normalized_value")
        def _control_value(self, value):
            if is_zoom_parameter(self.mapped_parameter):
                self.mapped_parameter.zoom(value * self._control_element.mapping_sensitivity)
            if self.mapped_parameter.is_quantized:
                steps = self._quantized_stepper.advance(value)
                if steps != 0:
                    self.mapped_parameter.value = self._clamp_value_to_parameter_range(
                        self.mapped_parameter.value + steps,
                    )
            else:
                value_offset = value * self._control_element.mapping_sensitivity
                self.mapped_parameter.linear_value = self._clamp_value_to_parameter_range(
                    self.mapped_parameter.linear_value + value_offset,
                )

        def _clamp_value_to_parameter_range(self, value):
            return clamp(value, self.mapped_parameter.min, self.mapped_parameter.max)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/scene.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 8199 bytes
import Live

from ...base import listens, liveobj_changed, liveobj_valid
from ..component import Component
from ..control import ButtonControl
from .clip_slot import ClipSlotComponent, find_nearest_color, is_button_pressed


class SceneComponent(Component):
    clip_slot_component_type = ClipSlotComponent
    launch_button = ButtonControl()

    def __init__(self, session_ring=None, *a, **k):
        self._controlled_tracks = []
        (super().__init__)(*a, **k)
        self._session_ring = session_ring
        self._scene = None
        self._clip_slots = []
        self._color_palette = None
        self._color_table = None
        for _ in range(session_ring.num_tracks):
            new_slot = self._create_clip_slot()
            self._clip_slots.append(new_slot)

        self._triggered_color = "Session.SceneTriggered"
        self._scene_color = "Session.Scene"
        self._no_scene_color = "Session.NoScene"
        self._track_offset = 0
        self._select_button = None
        self._delete_button = None
        self._duplicate_button = None
        self._SceneComponent__on_track_list_changed.subject = session_ring

    @listens("tracks")
    def __on_track_list_changed(self):
        self._update_controlled_tracks()

    def set_scene(self, scene):
        if liveobj_changed(scene, self._scene):
            self._scene = scene
            self._SceneComponent__on_is_triggered_changed.subject = scene
            self._SceneComponent__on_scene_color_changed.subject = scene
            self.update()

    def set_launch_button(self, button):
        self.launch_button.set_control_element(button)
        self.update()

    def set_select_button(self, button):
        self._select_button = button

    def set_delete_button(self, button):
        self._delete_button = button

    def set_duplicate_button(self, button):
        self._duplicate_button = button

    def set_track_offset(self, offset):
        if offset != self._track_offset:
            self._track_offset = offset
            self._update_controlled_tracks()

    def set_color_palette(self, palette):
        self._color_palette = palette

    def set_color_table(self, table):
        self._color_table = table

    def clip_slot(self, index):
        return self._clip_slots[index]

    def update(self):
        super().update()
        if liveobj_valid(self._scene) and self.is_enabled():
            clip_slots_to_use = self.build_clip_slot_list()
            for slot_wrapper, clip_slot in zip(self._clip_slots, clip_slots_to_use):
                slot_wrapper.set_clip_slot(clip_slot)

        else:
            for slot in self._clip_slots:
                slot.set_clip_slot(None)

        self._update_launch_button()

    def _update_controlled_tracks(self):
        controlled_tracks = self._session_ring.controlled_tracks()
        if controlled_tracks != self._controlled_tracks:
            self.update()
            self._controlled_tracks = controlled_tracks

    def _determine_actual_track_offset(self, tracks):
        actual_track_offset = self._track_offset
        if self._track_offset > 0:
            real_offset = 0
            visible_tracks = 0
            while visible_tracks < self._track_offset:
                if len(tracks) > real_offset:
                    if tracks[real_offset].is_visible:
                        visible_tracks += 1
                    else:
                        real_offset += 1

            actual_track_offset = real_offset
        return actual_track_offset

    def build_clip_slot_list(self):
        slots_to_use = []
        tracks = self.song.tracks
        track_offset = self._determine_actual_track_offset(tracks)
        clip_slots = self._scene.clip_slots
        for _ in self._clip_slots:
            while len(tracks) > track_offset:
                if not tracks[track_offset].is_visible:
                    track_offset += 1

            if len(clip_slots) > track_offset:
                slots_to_use.append(clip_slots[track_offset])
            else:
                slots_to_use.append(None)
            track_offset += 1

        return slots_to_use

    @launch_button.pressed
    def launch_button(self, value):
        self._on_launch_button_pressed()

    def _on_launch_button_pressed(self):
        if is_button_pressed(self._select_button):
            self._do_select_scene(self._scene)
        else:
            if liveobj_valid(self._scene):
                if is_button_pressed(self._duplicate_button):
                    self._do_duplicate_scene(self._scene)
                else:
                    if is_button_pressed(self._delete_button):
                        self._do_delete_scene(self._scene)
                    else:
                        self._do_launch_scene(True)

    @launch_button.released
    def launch_button(self, value):
        self._on_launch_button_released()

    def _on_launch_button_released(self):
        if not is_button_pressed(self._select_button):
            if liveobj_valid(self._scene):
                if not is_button_pressed(self._duplicate_button):
                    if not is_button_pressed(self._delete_button):
                        self._do_launch_scene(False)

    def _do_select_scene(self, scene_for_overrides):
        if liveobj_valid(self._scene):
            view = self.song.view
            if view.selected_scene != self._scene:
                view.selected_scene = self._scene
                self._on_scene_selected()

    def _on_scene_selected(self):
        pass

    def _do_delete_scene(self, scene_for_overrides):
        try:
            if liveobj_valid(self._scene):
                song = self.song
                song.delete_scene(list(song.scenes).index(self._scene))
                self._on_scene_deleted()
        except RuntimeError:
            pass

    def _on_scene_deleted(self):
        pass

    def _do_duplicate_scene(self, scene_for_overrides):
        try:
            song = self.song
            song.duplicate_scene(list(song.scenes).index(self._scene))
            self._on_scene_duplicated()
        except (Live.Base.LimitationError, IndexError, RuntimeError):
            pass

    def _on_scene_duplicated(self):
        pass

    def _do_launch_scene(self, value):
        launched = False
        if self.launch_button.is_momentary:
            self._scene.set_fire_button_state(value != 0)
            launched = value != 0
        else:
            if value != 0:
                self._scene.fire()
                launched = True
        if launched:
            if self.song.select_on_launch:
                self.song.view.selected_scene = self._scene

    @listens("is_triggered")
    def __on_is_triggered_changed(self):
        self._update_launch_button()

    @listens("color")
    def __on_scene_color_changed(self):
        self._update_launch_button()

    def _color_value(self, color):
        value = None
        if self._color_palette:
            value = self._color_palette.get(color, None)
        if value is None:
            if self._color_table:
                value = find_nearest_color(self._color_table, color)
        return value

    def _update_launch_button(self):
        if self.is_enabled():
            value_to_send = self._no_scene_color
            if liveobj_valid(self._scene):
                value_to_send = self._scene_color
                if self._scene.is_triggered:
                    value_to_send = self._triggered_color
                else:
                    possible_color = self._color_value(self._scene.color)
                    if possible_color:
                        value_to_send = possible_color
            self.launch_button.color = value_to_send

    def _create_clip_slot(self):
        return self.clip_slot_component_type(parent=self)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/components/clip_slot.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 8124 bytes
from ...base import const, depends, listens
from ...live import action, display_name, is_track_armed, liveobj_changed, liveobj_valid
from .. import Component
from ..controls import ButtonControl
from ..display import Renderable
from ..skin import LiveObjSkinEntry, OptionalSkinEntry


class ClipSlotComponent(Component, Renderable):
    launch_button = ButtonControl()
    select_button = ButtonControl(color=None)
    delete_button = ButtonControl(color=None)
    duplicate_button = ButtonControl(color=None)
    copy_button = ButtonControl(color=None)
    include_in_top_level_state = False

    @depends(clipboard=(const(None)))
    def __init__(self, clipboard=None, *a, **k):
        (super().__init__)(*a, **k)
        self._clipboard = clipboard
        self._clip_slot = None
        self._non_player_track = None

        def make_property_slot(property_name, update_method):
            return self.register_slot(None, update_method, property_name)

        self._track_property_slots = [
            make_property_slot("arm", self._update_launch_button_color),
            make_property_slot("implicit_arm", self._update_launch_button_color),
            make_property_slot("input_routing_type", self._update_launch_button_color),
        ]
        self._clip_slot_property_slots = [
            make_property_slot("color", self._update_launch_button_color),
            make_property_slot("playing_status", self._update_launch_button_color),
            make_property_slot("has_stop_button", self._update_launch_button_color),
            make_property_slot("has_clip", self._update_clip_property_slots),
            make_property_slot("controls_other_clips", self._update_clip_property_slots),
        ]
        self._clip_property_slots = [
            make_property_slot("color", self._update_launch_button_color),
            make_property_slot("playing_status", self._update_launch_button_color),
            make_property_slot("is_recording", self._update_launch_button_color),
        ]

    @property
    def clip_slot(self):
        return self._clip_slot

    def set_clip_slot(self, clip_slot):
        if liveobj_changed(clip_slot, self._clip_slot):
            self._clip_slot = clip_slot
            for slot in self._clip_slot_property_slots:
                slot.subject = clip_slot

            self._ClipSlotComponent__on_slot_triggered_changed.subject = clip_slot
            self._update_clip_property_slots(update_launch_button=False)
            track = clip_slot.canonical_parent if clip_slot else None
            if track:
                if track in self.song.tracks:
                    for slot in self._track_property_slots:
                        slot.subject = track

            self.update()

    def set_non_player_track(self, track):
        self.set_clip_slot(None)
        self._non_player_track = track

    def set_launch_button(self, button):
        self.launch_button.set_control_element(button)
        self.update()

    @launch_button.pressed
    def launch_button(self, _):
        self._on_launch_button_pressed()

    def _on_launch_button_pressed(self):
        slot_name = display_name(self._clip_slot) if liveobj_valid(self._clip_slot) else ""
        if self.select_button.is_pressed:
            if action.select(self._clip_slot):
                self.notify(self.notifications.Clip.select, slot_name)
            else:
                action.select(self._non_player_track)
        else:
            if self.duplicate_button.is_pressed:
                action.duplicate(self._clip_slot)
            else:
                if self._is_copying():
                    self._clipboard.copy_or_paste(self._clip_slot)
                else:
                    if self.delete_button.is_pressed:
                        if action.delete(self._clip_slot):
                            self.notify(self.notifications.Clip.delete, slot_name)
                        else:
                            self.notify(self.notifications.Clip.error_delete_empty_slot)
                    else:
                        self._do_launch_slot()

    def _do_launch_slot(self):
        action.fire((self._clip_slot), button_state=True)

    @launch_button.released
    def launch_button(self, _):
        self._on_launch_button_released()

    def _on_launch_button_released(self):
        if self.launch_button.is_momentary:
            if not self._any_modifier_pressed():
                if not self._is_copying():
                    action.fire((self._clip_slot), button_state=False)

    def _has_clip(self):
        return liveobj_valid(self._clip_slot) and self._clip_slot.has_clip

    def _is_copying(self):
        return (self.copy_button.is_pressed) or ((self._clipboard) and (self._clipboard.has_content))

    def _any_modifier_pressed(self):
        return (
            self.select_button.is_pressed
            or self.delete_button.is_pressed
            or self.duplicate_button.is_pressed
            or self.copy_button.is_pressed
        )

    def update(self):
        super().update()
        self._update_launch_button_color()

    def _update_launch_button_color(self):
        value_to_send = "Session.NoSlot"
        if liveobj_valid(self._clip_slot):
            track = self._clip_slot.canonical_parent
            slot_or_clip = self._clip_slot.clip if self._has_clip() else self._clip_slot
            value_to_send = self._feedback_value(track, slot_or_clip)
        self.launch_button.color = value_to_send

    def _feedback_value(self, track, slot_or_clip):
        is_clip = self._has_clip() or getattr(slot_or_clip, "controls_other_clips", False)
        value = "Session.Slot"
        if slot_or_clip.is_triggered:
            value = self._triggered_color(slot_or_clip, is_clip)
        else:
            if slot_or_clip.is_playing:
                value = "Session.Clip{}".format("Recording" if slot_or_clip.is_recording else "Playing")
            else:
                if is_clip:
                    value = "Session.ClipStopped"
                else:
                    if not self._clip_slot.has_stop_button:
                        value = OptionalSkinEntry("Session.SlotLacksStop", "Session.Slot")
                    else:
                        if is_track_armed(track):
                            if self._clip_slot.has_stop_button:
                                value = "Session.SlotRecordButton"
        return LiveObjSkinEntry(value, slot_or_clip)

    @staticmethod
    def _triggered_color(slot_or_clip, is_clip):
        rec_or_play = "Triggered{}".format("Record" if slot_or_clip.will_record_on_start else "Play")
        default_value = f"Session.Clip{rec_or_play}"
        if not is_clip:
            return OptionalSkinEntry(f"Session.Slot{rec_or_play}", default_value)
        return default_value

    def _update_clip_property_slots(self, update_launch_button=True):
        clip = self._clip_slot.clip if self._clip_slot else None
        for slot in self._clip_property_slots:
            slot.subject = clip

        if update_launch_button:
            self._update_launch_button_color()

    @listens("is_triggered")
    def __on_slot_triggered_changed(self):
        if not self._has_clip():
            self._update_launch_button_color()



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.MixerDevice import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class PlaybackMode:
    """
    Class that represent an enumeration of values for PlaybackMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    classic = 0
    one_shot = 1
    slicing = 2


class SimplerDevice:
    """
    This class represents a Simpler device.
    """

    class View:
        """
        Representing the view aspects of a simpler device.
        """

        def add_is_collapsed_listener(self, listener: Callable) -> None:
            """
            add_is_collapsed_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "is_collapsed" has changed.

                C++ signature :
                    void add_is_collapsed_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def add_sample_end_listener(self, listener: Callable) -> None:
            """
            add_sample_end_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_end" has changed.

                C++ signature :
                    void add_sample_end_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_sample_env_fade_in_listener(self, listener: Callable) -> None:
            """
            add_sample_env_fade_in_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_env_fade_in" has changed.

                C++ signature :
                    void add_sample_env_fade_in_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_sample_env_fade_out_listener(self, listener: Callable) -> None:
            """
            add_sample_env_fade_out_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_env_fade_out" has changed.

                C++ signature :
                    void add_sample_env_fade_out_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_sample_loop_end_listener(self, listener: Callable) -> None:
            """
            add_sample_loop_end_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_loop_end" has changed.

                C++ signature :
                    void add_sample_loop_end_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_sample_loop_fade_listener(self, listener: Callable) -> None:
            """
            add_sample_loop_fade_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_loop_fade" has changed.

                C++ signature :
                    void add_sample_loop_fade_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_sample_loop_start_listener(self, listener: Callable) -> None:
            """
            add_sample_loop_start_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_loop_start" has changed.

                C++ signature :
                    void add_sample_loop_start_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_sample_start_listener(self, listener: Callable) -> None:
            """
            add_sample_start_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "sample_start" has changed.

                C++ signature :
                    void add_sample_start_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def add_selected_slice_listener(self, listener: Callable) -> None:
            """
            add_selected_slice_listener( (View)arg1, (object)arg2) -> None :
                Add a listener function or method, which will be called as soon as the
                property "selected_slice" has changed.

                C++ signature :
                    void add_selected_slice_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def is_collapsed_has_listener(self, listener: Callable) -> bool:
            """
            is_collapsed_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "is_collapsed".

                C++ signature :
                    bool is_collapsed_has_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def remove_is_collapsed_listener(self, listener: Callable) -> None:
            """
            remove_is_collapsed_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "is_collapsed".

                C++ signature :
                    void remove_is_collapsed_listener(TPyViewData<ADevice>,boost::python::api::object)
            """

        def remove_sample_end_listener(self, listener: Callable) -> None:
            """
            remove_sample_end_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_end".

                C++ signature :
                    void remove_sample_end_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_sample_env_fade_in_listener(self, listener: Callable) -> None:
            """
            remove_sample_env_fade_in_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_env_fade_in".

                C++ signature :
                    void remove_sample_env_fade_in_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_sample_env_fade_out_listener(self, listener: Callable) -> None:
            """
            remove_sample_env_fade_out_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_env_fade_out".

                C++ signature :
                    void remove_sample_env_fade_out_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_sample_loop_end_listener(self, listener: Callable) -> None:
            """
            remove_sample_loop_end_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_loop_end".

                C++ signature :
                    void remove_sample_loop_end_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_sample_loop_fade_listener(self, listener: Callable) -> None:
            """
            remove_sample_loop_fade_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_loop_fade".

                C++ signature :
                    void remove_sample_loop_fade_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_sample_loop_start_listener(self, listener: Callable) -> None:
            """
            remove_sample_loop_start_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_loop_start".

                C++ signature :
                    void remove_sample_loop_start_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_sample_start_listener(self, listener: Callable) -> None:
            """
            remove_sample_start_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "sample_start".

                C++ signature :
                    void remove_sample_start_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def remove_selected_slice_listener(self, listener: Callable) -> None:
            """
            remove_selected_slice_listener( (View)arg1, (object)arg2) -> None :
                Remove a previously set listener function or method from
                property "selected_slice".

                C++ signature :
                    void remove_selected_slice_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_end_has_listener(self, listener: Callable) -> bool:
            """
            sample_end_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_end".

                C++ signature :
                    bool sample_end_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_env_fade_in_has_listener(self, listener: Callable) -> bool:
            """
            sample_env_fade_in_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_env_fade_in".

                C++ signature :
                    bool sample_env_fade_in_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_env_fade_out_has_listener(self, listener: Callable) -> bool:
            """
            sample_env_fade_out_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_env_fade_out".

                C++ signature :
                    bool sample_env_fade_out_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_loop_end_has_listener(self, listener: Callable) -> bool:
            """
            sample_loop_end_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_loop_end".

                C++ signature :
                    bool sample_loop_end_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_loop_fade_has_listener(self, listener: Callable) -> bool:
            """
            sample_loop_fade_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_loop_fade".

                C++ signature :
                    bool sample_loop_fade_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_loop_start_has_listener(self, listener: Callable) -> bool:
            """
            sample_loop_start_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_loop_start".

                C++ signature :
                    bool sample_loop_start_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def sample_start_has_listener(self, listener: Callable) -> bool:
            """
            sample_start_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "sample_start".

                C++ signature :
                    bool sample_start_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        def selected_slice_has_listener(self, listener: Callable) -> bool:
            """
            selected_slice_has_listener( (View)arg1, (object)arg2) -> bool :
                Returns true, if the given listener function or method is connected
                to the property "selected_slice".

                C++ signature :
                    bool selected_slice_has_listener(TSimplerDevicePyViewData,boost::python::api::object)
            """

        @property
        def canonical_parent(self) -> Any:
            """
            Get the canonical parent of the View.
            """

        @property
        def is_collapsed(self) -> bool:
            """
            Get/Set/Listen if the device is shown collapsed in the device chain.
            """

        @is_collapsed.setter
        def is_collapsed(self, value: bool) -> None:
            pass

        @property
        def sample_end(self) -> Any:
            """
            Access to the modulated samples end position in samples. Returns -1 in case there is no sample loaded.
            """

        @property
        def sample_env_fade_in(self) -> Any:
            """
            Access to the envelope fade-in time in samples. Returned value is only in use when Simpler is in one-shot mode. Returns -1 in case there is no sample loaded.
            """

        @property
        def sample_env_fade_out(self) -> Any:
            """
            Access to the envelope fade-out time in samples. Returned value is only in use when Simpler is in one-shot mode. Returns -1 in case there is no sample loaded.
            """

        @property
        def sample_loop_end(self) -> Any:
            """
            Access to the modulated samples loop end position in samples. Returns -1 in case there is no sample loaded.
            """

        @property
        def sample_loop_fade(self) -> Any:
            """
            Access to the modulated samples loop fade position in samples. Returns -1 in case there is no sample loaded.
            """

        @property
        def sample_loop_start(self) -> Any:
            """
            Access to the modulated samples loop start position in samples. Returns -1 in case there is no sample loaded.
            """

        @property
        def sample_start(self) -> Any:
            """
            Access to the modulated samples start position in samples. Returns -1 in case there is no sample loaded.
            """

        @property
        def selected_slice(self) -> Any:
            """
            Access to the selected slice.
            """

    def add_can_warp_as_listener(self, listener: Callable) -> None:
        """
        add_can_warp_as_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "can_warp_as" has changed.

            C++ signature :
                void add_can_warp_as_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_can_warp_double_listener(self, listener: Callable) -> None:
        """
        add_can_warp_double_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "can_warp_double" has changed.

            C++ signature :
                void add_can_warp_double_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_can_warp_half_listener(self, listener: Callable) -> None:
        """
        add_can_warp_half_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "can_warp_half" has changed.

            C++ signature :
                void add_can_warp_half_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_is_active_listener(self, listener: Callable) -> None:
        """
        add_is_active_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "is_active" has changed.

            C++ signature :
                void add_is_active_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_latency_in_ms_listener(self, listener: Callable) -> None:
        """
        add_latency_in_ms_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "latency_in_ms" has changed.

            C++ signature :
                void add_latency_in_ms_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_latency_in_samples_listener(self, listener: Callable) -> None:
        """
        add_latency_in_samples_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "latency_in_samples" has changed.

            C++ signature :
                void add_latency_in_samples_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_multi_sample_mode_listener(self, listener: Callable) -> None:
        """
        add_multi_sample_mode_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "multi_sample_mode" has changed.

            C++ signature :
                void add_multi_sample_mode_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_name_listener(self, listener: Callable) -> None:
        """
        add_name_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "name" has changed.

            C++ signature :
                void add_name_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_note_pitch_bend_range_listener(self, listener: Callable) -> None:
        """
        add_note_pitch_bend_range_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "note_pitch_bend_range" has changed.

            C++ signature :
                void add_note_pitch_bend_range_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_pad_slicing_listener(self, listener: Callable) -> None:
        """
        add_pad_slicing_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "pad_slicing" has changed.

            C++ signature :
                void add_pad_slicing_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_parameters_listener(self, listener: Callable) -> None:
        """
        add_parameters_listener( (Device)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "parameters" has changed.

            C++ signature :
                void add_parameters_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def add_pitch_bend_range_listener(self, listener: Callable) -> None:
        """
        add_pitch_bend_range_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "pitch_bend_range" has changed.

            C++ signature :
                void add_pitch_bend_range_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_playback_mode_listener(self, listener: Callable) -> None:
        """
        add_playback_mode_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "playback_mode" has changed.

            C++ signature :
                void add_playback_mode_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_playing_position_enabled_listener(self, listener: Callable) -> None:
        """
        add_playing_position_enabled_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "playing_position_enabled" has changed.

            C++ signature :
                void add_playing_position_enabled_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_playing_position_listener(self, listener: Callable) -> None:
        """
        add_playing_position_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "playing_position" has changed.

            C++ signature :
                void add_playing_position_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_retrigger_listener(self, listener: Callable) -> None:
        """
        add_retrigger_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "retrigger" has changed.

            C++ signature :
                void add_retrigger_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_sample_listener(self, listener: Callable) -> None:
        """
        add_sample_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "sample" has changed.

            C++ signature :
                void add_sample_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_slicing_playback_mode_listener(self, listener: Callable) -> None:
        """
        add_slicing_playback_mode_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "slicing_playback_mode" has changed.

            C++ signature :
                void add_slicing_playback_mode_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def add_voices_listener(self, listener: Callable) -> None:
        """
        add_voices_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "voices" has changed.

            C++ signature :
                void add_voices_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def can_warp_as_has_listener(self, listener: Callable) -> bool:
        """
        can_warp_as_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "can_warp_as".

            C++ signature :
                bool can_warp_as_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def can_warp_double_has_listener(self, listener: Callable) -> bool:
        """
        can_warp_double_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "can_warp_double".

            C++ signature :
                bool can_warp_double_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def can_warp_half_has_listener(self, listener: Callable) -> bool:
        """
        can_warp_half_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "can_warp_half".

            C++ signature :
                bool can_warp_half_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def crop(self) -> None:
        """
        crop( (SimplerDevice)self) -> None :
            Crop the loaded sample to the active area between start- and end marker.
            Calling this method on an empty simpler raises an error.

            C++ signature :
                void crop(TSimplerDevicePyHandle)
        """

    def guess_playback_length(self) -> float:
        """
        guess_playback_length( (SimplerDevice)self) -> float :
            Return an estimated beat time for the playback length between start- and end-marker.
            Calling this method on an empty simpler raises an error.

            C++ signature :
                double guess_playback_length(TSimplerDevicePyHandle)
        """

    def is_active_has_listener(self, listener: Callable) -> bool:
        """
        is_active_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "is_active".

            C++ signature :
                bool is_active_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def latency_in_ms_has_listener(self, listener: Callable) -> bool:
        """
        latency_in_ms_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "latency_in_ms".

            C++ signature :
                bool latency_in_ms_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def latency_in_samples_has_listener(self, listener: Callable) -> bool:
        """
        latency_in_samples_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "latency_in_samples".

            C++ signature :
                bool latency_in_samples_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def multi_sample_mode_has_listener(self, listener: Callable) -> bool:
        """
        multi_sample_mode_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "multi_sample_mode".

            C++ signature :
                bool multi_sample_mode_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def name_has_listener(self, listener: Callable) -> bool:
        """
        name_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "name".

            C++ signature :
                bool name_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def note_pitch_bend_range_has_listener(self, listener: Callable) -> bool:
        """
        note_pitch_bend_range_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "note_pitch_bend_range".

            C++ signature :
                bool note_pitch_bend_range_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def pad_slicing_has_listener(self, listener: Callable) -> bool:
        """
        pad_slicing_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "pad_slicing".

            C++ signature :
                bool pad_slicing_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def parameters_has_listener(self, listener: Callable) -> bool:
        """
        parameters_has_listener( (Device)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "parameters".

            C++ signature :
                bool parameters_has_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def pitch_bend_range_has_listener(self, listener: Callable) -> bool:
        """
        pitch_bend_range_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "pitch_bend_range".

            C++ signature :
                bool pitch_bend_range_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def playback_mode_has_listener(self, listener: Callable) -> bool:
        """
        playback_mode_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "playback_mode".

            C++ signature :
                bool playback_mode_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def playing_position_enabled_has_listener(self, listener: Callable) -> bool:
        """
        playing_position_enabled_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "playing_position_enabled".

            C++ signature :
                bool playing_position_enabled_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def playing_position_has_listener(self, listener: Callable) -> bool:
        """
        playing_position_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "playing_position".

            C++ signature :
                bool playing_position_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_can_warp_as_listener(self, listener: Callable) -> None:
        """
        remove_can_warp_as_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "can_warp_as".

            C++ signature :
                void remove_can_warp_as_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_can_warp_double_listener(self, listener: Callable) -> None:
        """
        remove_can_warp_double_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "can_warp_double".

            C++ signature :
                void remove_can_warp_double_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_can_warp_half_listener(self, listener: Callable) -> None:
        """
        remove_can_warp_half_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "can_warp_half".

            C++ signature :
                void remove_can_warp_half_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_is_active_listener(self, listener: Callable) -> None:
        """
        remove_is_active_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "is_active".

            C++ signature :
                void remove_is_active_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_latency_in_ms_listener(self, listener: Callable) -> None:
        """
        remove_latency_in_ms_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "latency_in_ms".

            C++ signature :
                void remove_latency_in_ms_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_latency_in_samples_listener(self, listener: Callable) -> None:
        """
        remove_latency_in_samples_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "latency_in_samples".

            C++ signature :
                void remove_latency_in_samples_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_multi_sample_mode_listener(self, listener: Callable) -> None:
        """
        remove_multi_sample_mode_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "multi_sample_mode".

            C++ signature :
                void remove_multi_sample_mode_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_name_listener(self, listener: Callable) -> None:
        """
        remove_name_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "name".

            C++ signature :
                void remove_name_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_note_pitch_bend_range_listener(self, listener: Callable) -> None:
        """
        remove_note_pitch_bend_range_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "note_pitch_bend_range".

            C++ signature :
                void remove_note_pitch_bend_range_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_pad_slicing_listener(self, listener: Callable) -> None:
        """
        remove_pad_slicing_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "pad_slicing".

            C++ signature :
                void remove_pad_slicing_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_parameters_listener(self, listener: Callable) -> None:
        """
        remove_parameters_listener( (Device)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "parameters".

            C++ signature :
                void remove_parameters_listener(TPyHandle<ADevice>,boost::python::api::object)
        """

    def remove_pitch_bend_range_listener(self, listener: Callable) -> None:
        """
        remove_pitch_bend_range_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "pitch_bend_range".

            C++ signature :
                void remove_pitch_bend_range_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_playback_mode_listener(self, listener: Callable) -> None:
        """
        remove_playback_mode_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "playback_mode".

            C++ signature :
                void remove_playback_mode_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_playing_position_enabled_listener(self, listener: Callable) -> None:
        """
        remove_playing_position_enabled_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "playing_position_enabled".

            C++ signature :
                void remove_playing_position_enabled_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_playing_position_listener(self, listener: Callable) -> None:
        """
        remove_playing_position_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "playing_position".

            C++ signature :
                void remove_playing_position_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_retrigger_listener(self, listener: Callable) -> None:
        """
        remove_retrigger_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "retrigger".

            C++ signature :
                void remove_retrigger_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_sample_listener(self, listener: Callable) -> None:
        """
        remove_sample_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "sample".

            C++ signature :
                void remove_sample_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_slicing_playback_mode_listener(self, listener: Callable) -> None:
        """
        remove_slicing_playback_mode_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "slicing_playback_mode".

            C++ signature :
                void remove_slicing_playback_mode_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def remove_voices_listener(self, listener: Callable) -> None:
        """
        remove_voices_listener( (SimplerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "voices".

            C++ signature :
                void remove_voices_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def retrigger_has_listener(self, listener: Callable) -> bool:
        """
        retrigger_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "retrigger".

            C++ signature :
                bool retrigger_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def reverse(self) -> None:
        """
        reverse( (SimplerDevice)self) -> None :
            Reverse the loaded sample.
            Calling this method on an empty simpler raises an error.

            C++ signature :
                void reverse(TSimplerDevicePyHandle)
        """

    def sample_has_listener(self, listener: Callable) -> bool:
        """
        sample_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "sample".

            C++ signature :
                bool sample_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def slicing_playback_mode_has_listener(self, listener: Callable) -> bool:
        """
        slicing_playback_mode_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "slicing_playback_mode".

            C++ signature :
                bool slicing_playback_mode_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def store_chosen_bank(self, arg2: int, arg3: int) -> None:
        """
        store_chosen_bank( (Device)arg1, (int)arg2, (int)arg3) -> None :
            Set the selected bank in the device for persistency.

            C++ signature :
                void store_chosen_bank(TPyHandle<ADevice>,int,int)
        """

    def voices_has_listener(self, listener: Callable) -> bool:
        """
        voices_has_listener( (SimplerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "voices".

            C++ signature :
                bool voices_has_listener(TSimplerDevicePyHandle,boost::python::api::object)
        """

    def warp_as(self, beat_time: float) -> None:
        """
        warp_as( (SimplerDevice)self, (float)beat_time) -> None :
            Warp the playback region between start- and end-marker as the given length.
            Calling this method on an empty simpler raises an error.

            C++ signature :
                void warp_as(TSimplerDevicePyHandle,double)
        """

    def warp_double(self) -> None:
        """
        warp_double( (SimplerDevice)self) -> None :
            Doubles the tempo for region between start- and end-marker.

            C++ signature :
                void warp_double(TSimplerDevicePyHandle)
        """

    def warp_half(self) -> None:
        """
        warp_half( (SimplerDevice)self) -> None :
            Halves the tempo for region between start- and end-marker.

            C++ signature :
                void warp_half(TSimplerDevicePyHandle)
        """

    @property
    def can_have_chains(self) -> bool:
        """
        Returns true if the device is a rack.
        """

    @property
    def can_have_drum_pads(self) -> bool:
        """
        Returns true if the device is a drum rack.
        """

    @property
    def can_warp_as(self) -> bool:
        """
        Returns true if warp_as is available.
        """

    @property
    def can_warp_double(self) -> bool:
        """
        Returns true if warp_double is available.
        """

    @property
    def can_warp_half(self) -> bool:
        """
        Returns true if warp_half is available.
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the Device.
        """

    @property
    def class_display_name(self) -> str:
        """
        Return const access to the name of the device's class name as displayed in Live's browser and device chain
        """

    @property
    def class_name(self) -> str:
        """
        Return const access to the name of the device's class.
        """

    @property
    def is_active(self) -> bool:
        """
        Return const access to whether this device is active. This will be false bothwhen the device is off and when it's inside a rack device which is off.
        """

    @property
    def latency_in_ms(self) -> Any:
        """
        Returns the latency of the device in ms.
        """

    @property
    def latency_in_samples(self) -> Any:
        """
        Returns the latency of the device in samples.
        """

    @property
    def multi_sample_mode(self) -> Any:
        """
        Returns whether Simpler is in mulit-sample mode.
        """

    @property
    def name(self) -> str:
        """
        Return access to the name of the device.
        """

    @property
    def note_pitch_bend_range(self) -> Any:
        """
        Access to the Note Pitch Bend Range in Simpler.
        """

    @property
    def pad_slicing(self) -> Any:
        """
        When set to true, slices can be added in slicing mode by playing notes
        .that are not assigned to slices, yet.
        """

    @property
    def parameters(self) -> list[DeviceParameter]:
        """
        Const access to the list of available automatable parameters for this device.
        """

    @property
    def pitch_bend_range(self) -> Any:
        """
        Access to the Pitch Bend Range in Simpler.
        """

    @property
    def playback_mode(self) -> Any:
        """
        Access to Simpler's playback mode.
        """

    @property
    def playing_position(self) -> Any:
        """
        Constant access to the current playing position in the sample.
        The returned value is the normalized position between sample start and end.
        """

    @property
    def playing_position_enabled(self) -> Any:
        """
        Returns whether Simpler is showing the playing position.
        The returned value is True while the sample is played back
        """

    @property
    def retrigger(self) -> Any:
        """
        Access to Simpler's retrigger mode.
        """

    @property
    def sample(self) -> Any:
        """
        Get the loaded Sample.
        """

    @property
    def slicing_playback_mode(self) -> Any:
        """
        Access to Simpler's slicing playback mode.
        """

    @property
    def type(self) -> Any:
        """
        Return the type of the device.
        """

    @property
    def view(self) -> View:
        """
        Representing the view aspects of a device.
        """

    @property
    def voices(self) -> Any:
        """
        Access to the number of voices in Simpler.
        """


class SlicingPlaybackMode:
    """
    Class that represent an enumeration of values for SlicingPlaybackMode
    int([x]) -> integer
    int(x, base=10) -> integer

    Convert a number or string to an integer, or return 0 if no arguments
    are given.  If x is a number, return x.__int__().  For floating point
    numbers, this truncates towards zero.

    If x is not a number or if base is given, then x must be a string,
    bytes, or bytearray instance representing an integer literal in the
    given base.  The literal can be preceded by '+' or '-' and be surrounded
    by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
    Base 0 means to interpret the base from the string as an integer literal.
    >>> int('0b100', base=0)
    4
    """

    mono = 0
    poly = 1
    thru = 2


def get_available_voice_numbers() -> IntVector:
    """
    get_available_voice_numbers() -> IntVector :
        Get a vector of valid Simpler voice numbers.

        C++ signature :
            std::__1::vector<int, std::__1::allocator<int>> get_available_voice_numbers()
    """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v3/control_surface/controls/mapped.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 4608 bytes
from ableton.v2.control_surface.control import MappedSensitivitySettingControl as MappedSensitivitySettingControlBase

from ...base import EventObject, listens
from ...live import action, liveobj_valid
from .. import ABSOLUTE_MAP_MODES, EnumWrappingParameter
from . import ButtonControl as ButtonControlBase
from . import is_internal_parameter


class MappableButton(EventObject):
    def __init__(self, *a, **k):
        (super().__init__)(*a, **k)
        self._parameter = None

    def disconnect(self):
        self._parameter = None
        super().disconnect()

    @property
    def mapped_parameter(self):
        return self._parameter

    @mapped_parameter.setter
    def mapped_parameter(self, parameter):
        self._parameter = parameter if liveobj_valid(parameter) else None
        self.enabled = self._parameter is not None
        self._MappableButton__on_parameter_value_changed.subject = self._parameter
        self._MappableButton__on_parameter_value_changed()

    @listens("value")
    def __on_parameter_value_changed(self):
        self.is_on = liveobj_valid(self._parameter) and self._parameter.value


class MappedButtonControl(ButtonControlBase):
    class State(ButtonControlBase.State, MappableButton):
        def __init__(self, *a, **k):
            (super().__init__)(*a, **k)
            self.enabled = False

        def _call_listener(self, listener_name, *_):
            if listener_name == "pressed":
                action.toggle_or_cycle_parameter_value(self.mapped_parameter)


class MappedSensitivitySettingControl(MappedSensitivitySettingControlBase):
    class State(MappedSensitivitySettingControlBase.State):
        def __init__(self, default_sensitivity=None, fine_sensitivity=None, *a, **k):
            (super().__init__)(*a, **k)
            self.default_sensitivity = default_sensitivity or self.default_sensitivity
            self.fine_sensitivity = fine_sensitivity or self.fine_sensitivity

        def _update_direct_connection(self):
            self._control_value.subject = None
            self._absolute_control_value.subject = None
            self._quantized_stepper.reset()
            if self._control_element:
                if is_internal_parameter(self.mapped_parameter):
                    self._connect_internal_parameter()
                else:
                    self._update_control_element()

        @staticmethod
        def _is_parameter_valid(parameter):
            return not is_internal_parameter(parameter) or isinstance(parameter, EnumWrappingParameter)

        def _connect_internal_parameter(self):
            self._control_element.release_parameter()
            self._control_element.connect_to(self.mapped_parameter)
            if self._control_element.message_map_mode() in ABSOLUTE_MAP_MODES:
                self._absolute_control_value.subject = self._control_element
            else:
                self._control_value.subject = self._control_element
            self._update_control_sensitivity()

        @listens("value")
        def _absolute_control_value(self, value):
            step_size = self._control_element.max_value() / (self.mapped_parameter.max + 1)
            self.mapped_parameter.value = int(value / step_size)



from __future__ import annotations

from typing import Any, Callable

from Live import *
from Live.Application import *
from Live.Base import *
from Live.Browser import *
from Live.CcControlDevice import *
from Live.Chain import *
from Live.ChainMixerDevice import *
from Live.Clip import *
from Live.ClipSlot import *
from Live.CompressorDevice import *
from Live.Conversions import *
from Live.Device import *
from Live.DeviceIO import *
from Live.DeviceParameter import *
from Live.DriftDevice import *
from Live.DrumChain import *
from Live.DrumPad import *
from Live.Eq8Device import *
from Live.Groove import *
from Live.GroovePool import *
from Live.HybridReverbDevice import *
from Live.Listener import *
from Live.LomObject import *
from Live.MaxDevice import *
from Live.MeldDevice import *
from Live.MidiMap import *
from Live.PluginDevice import *
from Live.RackDevice import *
from Live.RoarDevice import *
from Live.Sample import *
from Live.Scene import *
from Live.ShifterDevice import *
from Live.SimplerDevice import *
from Live.Song import *
from Live.SpectralResonatorDevice import *
from Live.Track import *
from Live.WavetableDevice import *


class MixerDevice:
    """
    This class represents a Mixer Device in Live, which gives you
    access to the Volume and Panning properties of a Track.
    """

    class crossfade_assignments:
        """
        Class that represent an enumeration of values for crossfade_assignments
        int([x]) -> integer
        int(x, base=10) -> integer

        Convert a number or string to an integer, or return 0 if no arguments
        are given.  If x is a number, return x.__int__().  For floating point
        numbers, this truncates towards zero.

        If x is not a number or if base is given, then x must be a string,
        bytes, or bytearray instance representing an integer literal in the
        given base.  The literal can be preceded by '+' or '-' and be surrounded
        by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
        Base 0 means to interpret the base from the string as an integer literal.
        >>> int('0b100', base=0)
        4
        """

        A = 0
        NONE = 1
        B = 2

    class panning_modes:
        """
        Class that represent an enumeration of values for panning_modes
        int([x]) -> integer
        int(x, base=10) -> integer

        Convert a number or string to an integer, or return 0 if no arguments
        are given.  If x is a number, return x.__int__().  For floating point
        numbers, this truncates towards zero.

        If x is not a number or if base is given, then x must be a string,
        bytes, or bytearray instance representing an integer literal in the
        given base.  The literal can be preceded by '+' or '-' and be surrounded
        by whitespace.  The base defaults to 10.  Valid bases are 0 and 2-36.
        Base 0 means to interpret the base from the string as an integer literal.
        >>> int('0b100', base=0)
        4
        """

        stereo = 0
        stereo_split = 1

    def add_crossfade_assign_listener(self, listener: Callable) -> None:
        """
        add_crossfade_assign_listener( (MixerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "crossfade_assign" has changed.

            C++ signature :
                void add_crossfade_assign_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def add_panning_mode_listener(self, listener: Callable) -> None:
        """
        add_panning_mode_listener( (MixerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "panning_mode" has changed.

            C++ signature :
                void add_panning_mode_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def add_sends_listener(self, listener: Callable) -> None:
        """
        add_sends_listener( (MixerDevice)arg1, (object)arg2) -> None :
            Add a listener function or method, which will be called as soon as the
            property "sends" has changed.

            C++ signature :
                void add_sends_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def crossfade_assign_has_listener(self, listener: Callable) -> bool:
        """
        crossfade_assign_has_listener( (MixerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "crossfade_assign".

            C++ signature :
                bool crossfade_assign_has_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def panning_mode_has_listener(self, listener: Callable) -> bool:
        """
        panning_mode_has_listener( (MixerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "panning_mode".

            C++ signature :
                bool panning_mode_has_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def remove_crossfade_assign_listener(self, listener: Callable) -> None:
        """
        remove_crossfade_assign_listener( (MixerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "crossfade_assign".

            C++ signature :
                void remove_crossfade_assign_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def remove_panning_mode_listener(self, listener: Callable) -> None:
        """
        remove_panning_mode_listener( (MixerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "panning_mode".

            C++ signature :
                void remove_panning_mode_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def remove_sends_listener(self, listener: Callable) -> None:
        """
        remove_sends_listener( (MixerDevice)arg1, (object)arg2) -> None :
            Remove a previously set listener function or method from
            property "sends".

            C++ signature :
                void remove_sends_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    def sends_has_listener(self, listener: Callable) -> bool:
        """
        sends_has_listener( (MixerDevice)arg1, (object)arg2) -> bool :
            Returns true, if the given listener function or method is connected
            to the property "sends".

            C++ signature :
                bool sends_has_listener(TPyHandle<ATrackDevice>,boost::python::api::object)
        """

    @property
    def canonical_parent(self) -> Any:
        """
        Get the canonical parent of the mixer device.
        """

    @property
    def crossfade_assign(self) -> Any:
        """
        Player- and ReturnTracks only: Access to the Track's Crossfade Assign State.
        """

    @property
    def crossfader(self) -> Any:
        """
        MainTrack only: Const access to the Crossfader.
        """

    @property
    def cue_volume(self) -> Any:
        """
        MainTrack only: Const access to the Cue Volume Parameter.
        """

    @property
    def left_split_stereo(self) -> Any:
        """
        Const access to the Track's Left Split Stereo Panning Device Parameter.
        """

    @property
    def panning(self) -> Any:
        """
        Const access to the Tracks Panning Device Parameter.
        """

    @property
    def panning_mode(self) -> Any:
        """
        Access to the Track's Panning Mode.
        """

    @property
    def right_split_stereo(self) -> Any:
        """
        Const access to the Track's Right Split Stereo Panning Device Parameter.
        """

    @property
    def sends(self) -> Any:
        """
        Const access to the Tracks list of Send Amount Device Parameters.
        """

    @property
    def song_tempo(self) -> Any:
        """
        MainTrack only: Const access to the Song's Tempo.
        """

    @property
    def track_activator(self) -> Any:
        """
        Const access to the Tracks Activator Device Parameter.
        """

    @property
    def volume(self) -> Any:
        """
        Const access to the Tracks Volume Device Parameter.
        """



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/elements/display_data_source.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 3494 bytes
from functools import partial


def adjust_string_crop(original, length):
    length = int(length)
    return original[:length].ljust(length)


def adjust_string(original, length):
    length = int(length)
    resulting_string = original
    if len(resulting_string) > length:
        unit_db = resulting_string.endswith("dB") and resulting_string.find(".") != -1
        if len(resulting_string.strip()) > length:
            if unit_db:
                resulting_string = resulting_string[:-2]
        if len(resulting_string) > length:
            for char in (" ", "_", "i", "o", "u", "e", "a"):
                offset = 0 if char == " " else 1
                while len(resulting_string) > length:
                    if resulting_string.rfind(char, offset) > 0:
                        char_pos = resulting_string.rfind(char, offset)
                        resulting_string = resulting_string[:char_pos] + resulting_string[char_pos + 1 :]

            resulting_string = resulting_string[:length]
    if len(resulting_string) < length:
        resulting_string = resulting_string.ljust(length)
    return resulting_string


class DisplayDataSource:
    _separator = ""
    _adjust_string_fn = partial(adjust_string)

    def __init__(self, display_string="", separator=None, adjust_string_fn=adjust_string, *a, **k):
        (super().__init__)(*a, **k)
        if adjust_string_fn is not None:
            self._adjust_string_fn = partial(adjust_string_fn)
        if separator is not None:
            self._separator = separator
        self._display_string = display_string
        self._update_callback = None
        self._in_update = False

    @property
    def separator(self):
        return self._separator

    @separator.setter
    def separator(self, separator):
        if separator != self._separator:
            self._separator = separator
            self.update()

    def set_update_callback(self, update_callback):
        self._update_callback = update_callback
        if update_callback:
            self.update()

    def set_display_string(self, new_string):
        if self._display_string != new_string:
            self._display_string = new_string
            self.update()

    def clear(self):
        self.set_display_string("")
        self.separator = ""

    def update(self):
        self._in_update = True
        if self._update_callback is not None:
            self._update_callback()
        self._in_update = False

    def display_string(self):
        return self._display_string

    def adjust_string(self, width):
        return self._adjust_string_fn(self.display_string(), width)



# decompyle3 version 3.9.0
# Python bytecode version base 3.7.0 (3394)
# Decompiled from: Python 3.7.16 (default, Jan 17 2023, 09:28:58)
# [Clang 14.0.6 ]
# Embedded file name: output/Live/mac_universal_64_static/Release/python-bundle/MIDI Remote Scripts/ableton/v2/control_surface/components/device_navigation.py
# Compiled at: 2023-11-21 10:21:18
# Size of source mod 2**32: 7466 bytes
from itertools import chain

from ...base import SlotGroup, index_if, listens, liveobj_changed, liveobj_valid
from ...control_surface import device_to_appoint
from .item_lister import ItemListerComponent, ItemProvider


def is_empty_rack(rack):
    return rack.can_have_chains and len(rack.chains) == 0


def nested_device_parent(device):
    if device.can_have_chains:
        if device.view.is_showing_chain_devices:
            if not device.view.is_collapsed:
                return device.view.selected_chain
            return None
        return None
    return None


def collect_devices(track_or_chain, nesting_level=0):
    chain_devices = track_or_chain.devices if liveobj_valid(track_or_chain) else []
    devices = []
    for device in chain_devices:
        devices.append((device, nesting_level))
        devices.extend(collect_devices((nested_device_parent(device)), nesting_level=(nesting_level + 1)))

    return devices


class FlattenedDeviceChain(ItemProvider):
    def __init__(self, collect_devices_fun=collect_devices, *a, **k):
        (super().__init__)(*a, **k)
        self._device_parent = None
        self._devices = []
        self._selected_item = None
        self._collect_devices = collect_devices_fun

        def make_slot_group(event):
            slot_group = SlotGroup(self._update_devices, event)
            return self.register_disconnectable(slot_group)

        self._devices_changed = make_slot_group("devices")
        self._selected_chain_changed = make_slot_group("selected_chain")
        self._selected_pad_changed = make_slot_group("selected_drum_pad")
        self._collapsed_state_changed = make_slot_group("is_collapsed")
        self._macros_mapped_changed = make_slot_group("macros_mapped")
        self._chain_devices_visibility_changed = make_slot_group("is_showing_chain_devices")

    @property
    def items(self):
        return self._devices

    @property
    def selected_item(self):
        return self._selected_item

    @selected_item.setter
    def selected_item(self, device):
        if liveobj_changed(self._selected_item, device):
            self._selected_item = device
            self.notify_selected_item()

    @property
    def has_invalid_selection(self):
        return not liveobj_valid(self._selected_item)

    def set_device_parent(self, parent):
        self._device_parent = parent
        self._update_devices()

    def _update_devices(self, *_):
        self._devices = self._collect_devices(self._device_parent)
        self._update_listeners()
        self.notify_items()

    def _update_listeners(self):
        def get_rack_views(racks):
            return [rack.view for rack in racks]

        racks = [rack for rack, nesting_level in self._devices if getattr(rack, "can_have_chains", False)]
        rack_views = get_rack_views(racks)
        device_parents = chain((x.selected_chain for x in rack_views), [self._device_parent])

        def is_empty_pad_drum_rack(rack):
            return (
                rack.can_have_drum_pads and rack.view.selected_drum_pad and len(rack.view.selected_drum_pad.chains) == 0
            )

        empty_pad_drum_rack_views = get_rack_views(list(filter(is_empty_pad_drum_rack, racks)))
        self._devices_changed.replace_subjects(device_parents)
        self._selected_chain_changed.replace_subjects(rack_views)
        self._collapsed_state_changed.replace_subjects(rack_views)
        self._chain_devices_visibility_changed.replace_subjects(rack_views)
        self._macros_mapped_changed.replace_subjects(racks)
        self._selected_pad_changed.replace_subjects(empty_pad_drum_rack_views)


class DeviceNavigationComponent(ItemListerComponent):
    def __init__(self, device_component=None, item_provider=None, *a, **k):
        self._flattened_chain = item_provider or FlattenedDeviceChain()
        (super().__init__)(a, item_provider=self._flattened_chain, **k)
        self._device_component = device_component
        self._DeviceNavigationComponent__on_device_changed.subject = device_component
        self._last_pressed_button_index = -1
        self._selected_on_previous_press = None
        self.register_disconnectable(self._flattened_chain)
        self._on_selected_track_changed()
        self._on_selected_track_changed.subject = self.song.view
        self._DeviceNavigationComponent__on_device_changed()
        self._update_button_colors()

    @property
    def selected_item(self):
        return self.item_provider.selected_item

    def _on_select_button_pressed(self, button):
        self._select_item(self.items[button.index].item)

    def _show_selected_item(self):
        selected_item = self.item_provider.selected_item
        if selected_item is not None:
            items = self.item_provider.items
            if len(items) > self._num_visible_items:
                selected_index = index_if(lambda i: i[0] == selected_item, items)
                if selected_index >= self._num_visible_items + self.item_offset - 1 and selected_index < len(items) - 1:
                    self.item_offset = selected_index - self._num_visible_items + 2
                else:
                    if selected_index > 0:
                        if selected_index <= self.item_offset:
                            self.item_offset = selected_index - 1

    def _current_track(self):
        return self.song.view.selected_track

    @listens("selected_track")
    def _on_selected_track_changed(self):
        self._update_selected_track()

    def _update_selected_track(self):
        self._selected_track = self.song.view.selected_track
        selected_track = self._current_track()
        self.reset_offset()
        self._flattened_chain.set_device_parent(selected_track)
        self._device_selection_in_track_changed.subject = selected_track.view
        self._restore_selection(selected_track)

    def _restore_selection(self, selected_track):
        to_select = selected_track.view.selected_device
        self._select_item(to_select)

    def _select_item(self, device_or_pad):
        if device_or_pad:
            self._do_select_item(device_or_pad)
        self._update_item_provider(device_or_pad)

    def _do_select_item(self, device):
        appointed_device = device_to_appoint(device)
        self._appoint_device(appointed_device)
        self.song.view.select_device(device, False)
        self.song.appointed_device = appointed_device

    def _appoint_device(self, device):
        self._device_component.set_device(device)

    @listens("device")
    def __on_device_changed(self):
        self._update_device()

    def _update_device(self):
        self._update_item_provider(self._device_component.device())

    @listens("selected_device")
    def _device_selection_in_track_changed(self):
        new_selection = self.song.view.selected_track.view.selected_device
        self._update_item_provider(new_selection)

    def _update_item_provider(self, selection):
        self._flattened_chain.selected_item = selection
